import numpy
import numpy.testing as nptest
import pandas
import pytest
import seaborn
from matplotlib import pyplot

from wqio import utils, viz
from wqio.tests import helpers

BASELINE_IMAGES = "_baseline_images/viz_tests"
TOLERANCE = helpers.get_img_tolerance()


@pytest.fixture
def plot_data():
    data = numpy.array(
        [
            3.113,
            3.606,
            4.046,
            4.046,
            4.710,
            6.140,
            6.978,
            2.000,
            4.200,
            4.620,
            5.570,
            5.660,
            5.860,
            6.650,
            6.780,
            6.790,
            7.500,
            7.500,
            7.500,
            8.630,
            8.710,
            8.990,
            9.850,
            10.820,
            11.250,
            11.250,
            12.200,
            14.920,
            16.770,
            17.810,
            19.160,
            19.190,
            19.640,
            20.180,
            22.970,
        ]
    )
    return data


@pytest.fixture
def boxplot_data(plot_data):
    bs = [
        {
            "cihi": 10.82,
            "cilo": 6.1399999999999997,
            "fliers": numpy.array([22.97]),
            "iqr": 6.1099999999999994,
            "mean": 9.5888285714285733,
            "med": 7.5,
            "q1": 5.6150000000000002,
            "q3": 11.725,
            "whishi": 20.18,
            "whislo": 2.0,
        }
    ]
    return bs


@pytest.fixture
def whisk_flier_data():
    data = [
        2.20e-01,
        2.70e-01,
        3.08e-01,
        3.20e-01,
        4.10e-01,
        4.44e-01,
        4.82e-01,
        5.46e-01,
        6.05e-01,
        6.61e-01,
        7.16e-01,
        7.70e-01,
        8.24e-01,
        1.00e-03,
        4.90e-02,
        5.60e-02,
        1.40e-01,
        1.69e-01,
        1.83e-01,
        2.06e-01,
        2.10e-01,
        2.13e-01,
        2.86e-01,
        3.16e-01,
        3.40e-01,
        3.57e-01,
        3.71e-01,
        3.72e-01,
        3.78e-01,
        3.81e-01,
        3.86e-01,
        3.89e-01,
        3.90e-01,
        3.93e-01,
        4.00e-01,
        4.03e-01,
        4.10e-01,
        4.10e-01,
        4.29e-01,
        4.40e-01,
        4.40e-01,
        4.40e-01,
        4.46e-01,
        4.46e-01,
        4.50e-01,
        4.51e-01,
        4.52e-01,
        4.56e-01,
        4.60e-01,
        4.66e-01,
        4.72e-01,
        4.78e-01,
        4.81e-01,
        4.83e-01,
        4.86e-01,
        4.89e-01,
        4.98e-01,
        5.00e-01,
        5.00e-01,
        5.03e-01,
        5.18e-01,
        5.32e-01,
        5.37e-01,
        5.38e-01,
        5.68e-01,
        5.69e-01,
        5.78e-01,
        5.88e-01,
        5.94e-01,
        5.96e-01,
        6.02e-01,
        6.10e-01,
        6.10e-01,
        6.10e-01,
        6.19e-01,
        6.20e-01,
        6.20e-01,
        6.28e-01,
        6.38e-01,
        6.39e-01,
        6.42e-01,
        6.61e-01,
        6.71e-01,
        6.75e-01,
        6.80e-01,
        6.96e-01,
        7.00e-01,
        7.01e-01,
        7.09e-01,
        7.16e-01,
        7.17e-01,
        7.30e-01,
        7.62e-01,
        7.64e-01,
        7.69e-01,
        7.70e-01,
        7.77e-01,
        7.80e-01,
        8.06e-01,
        8.10e-01,
        8.23e-01,
        8.30e-01,
        8.50e-01,
        8.50e-01,
        8.56e-01,
        8.56e-01,
        8.80e-01,
        8.80e-01,
        8.93e-01,
        8.96e-01,
        8.97e-01,
        8.99e-01,
        9.22e-01,
        9.28e-01,
        9.30e-01,
        9.64e-01,
        9.65e-01,
        9.76e-01,
        9.79e-01,
        9.90e-01,
        9.99e-01,
        1.00e00,
        1.00e00,
        1.01e00,
        1.02e00,
        1.03e00,
        1.03e00,
        1.03e00,
        1.04e00,
        1.05e00,
        1.05e00,
        1.05e00,
        1.06e00,
        1.07e00,
        1.08e00,
        1.08e00,
        1.10e00,
        1.10e00,
        1.11e00,
        1.12e00,
        1.12e00,
        1.13e00,
        1.14e00,
        1.14e00,
        1.14e00,
        1.15e00,
        1.16e00,
        1.17e00,
        1.17e00,
        1.17e00,
        1.19e00,
        1.19e00,
        1.20e00,
        1.20e00,
        1.21e00,
        1.22e00,
        1.22e00,
        1.23e00,
        1.23e00,
        1.23e00,
        1.25e00,
        1.25e00,
        1.26e00,
        1.26e00,
        1.27e00,
        1.27e00,
        1.28e00,
        1.29e00,
        1.29e00,
        1.30e00,
        1.30e00,
        1.30e00,
        1.31e00,
        1.31e00,
        1.31e00,
        1.32e00,
        1.33e00,
        1.34e00,
        1.35e00,
        1.35e00,
        1.35e00,
        1.36e00,
        1.36e00,
        1.36e00,
        1.36e00,
        1.37e00,
        1.38e00,
        1.39e00,
        1.39e00,
        1.40e00,
        1.41e00,
        1.43e00,
        1.44e00,
        1.44e00,
        1.47e00,
        1.47e00,
        1.48e00,
        1.51e00,
        1.51e00,
        1.53e00,
        1.55e00,
        1.55e00,
        1.55e00,
        1.57e00,
        1.57e00,
        1.57e00,
        1.59e00,
        1.59e00,
        1.60e00,
        1.60e00,
        1.61e00,
        1.62e00,
        1.62e00,
        1.62e00,
        1.62e00,
        1.63e00,
        1.63e00,
        1.63e00,
        1.64e00,
        1.66e00,
        1.68e00,
        1.68e00,
        1.68e00,
        1.68e00,
        1.70e00,
        1.70e00,
        1.71e00,
        1.71e00,
        1.71e00,
        1.74e00,
        1.75e00,
        1.75e00,
        1.75e00,
        1.76e00,
        1.76e00,
        1.77e00,
        1.77e00,
        1.77e00,
        1.78e00,
        1.78e00,
        1.79e00,
        1.79e00,
        1.80e00,
        1.81e00,
        1.81e00,
        1.82e00,
        1.82e00,
        1.82e00,
        1.83e00,
        1.85e00,
        1.85e00,
        1.85e00,
        1.85e00,
        1.86e00,
        1.86e00,
        1.86e00,
        1.86e00,
        1.87e00,
        1.87e00,
        1.89e00,
        1.90e00,
        1.91e00,
        1.92e00,
        1.92e00,
        1.92e00,
        1.94e00,
        1.95e00,
        1.95e00,
        1.95e00,
        1.96e00,
        1.96e00,
        1.97e00,
        1.97e00,
        1.97e00,
        1.97e00,
        1.98e00,
        1.99e00,
        1.99e00,
        1.99e00,
        2.00e00,
        2.00e00,
        2.00e00,
        2.01e00,
        2.01e00,
        2.01e00,
        2.02e00,
        2.04e00,
        2.05e00,
        2.06e00,
        2.06e00,
        2.06e00,
        2.07e00,
        2.08e00,
        2.09e00,
        2.09e00,
        2.10e00,
        2.10e00,
        2.11e00,
        2.11e00,
        2.12e00,
        2.12e00,
        2.12e00,
        2.13e00,
        2.13e00,
        2.13e00,
        2.14e00,
        2.14e00,
        2.14e00,
        2.14e00,
        2.14e00,
        2.15e00,
        2.16e00,
        2.17e00,
        2.18e00,
        2.18e00,
        2.18e00,
        2.19e00,
        2.19e00,
        2.19e00,
        2.19e00,
        2.19e00,
        2.21e00,
        2.23e00,
        2.23e00,
        2.23e00,
        2.25e00,
        2.25e00,
        2.25e00,
        2.25e00,
        2.26e00,
        2.26e00,
        2.26e00,
        2.26e00,
        2.26e00,
        2.27e00,
        2.27e00,
        2.28e00,
        2.28e00,
        2.28e00,
        2.29e00,
        2.29e00,
        2.29e00,
        2.30e00,
        2.31e00,
        2.32e00,
        2.33e00,
        2.33e00,
        2.33e00,
        2.33e00,
        2.34e00,
        2.36e00,
        2.38e00,
        2.38e00,
        2.39e00,
        2.39e00,
        2.39e00,
        2.41e00,
        2.42e00,
        2.43e00,
        2.45e00,
        2.45e00,
        2.47e00,
        2.48e00,
        2.49e00,
        2.49e00,
        2.49e00,
        2.50e00,
        2.51e00,
        2.51e00,
        2.52e00,
        2.53e00,
        2.53e00,
        2.54e00,
        2.54e00,
        2.56e00,
        2.58e00,
        2.59e00,
        2.59e00,
        2.60e00,
        2.61e00,
        2.61e00,
        2.61e00,
        2.62e00,
        2.62e00,
        2.63e00,
        2.65e00,
        2.65e00,
        2.66e00,
        2.66e00,
        2.68e00,
        2.69e00,
        2.69e00,
        2.70e00,
        2.72e00,
        2.72e00,
        2.73e00,
        2.75e00,
        2.77e00,
        2.78e00,
        2.79e00,
        2.81e00,
        2.81e00,
        2.82e00,
        2.84e00,
        2.84e00,
        2.85e00,
        2.85e00,
        2.86e00,
        2.86e00,
        2.88e00,
        2.92e00,
        2.93e00,
        2.93e00,
        2.95e00,
        2.96e00,
        2.96e00,
        2.99e00,
        3.00e00,
        3.01e00,
        3.02e00,
        3.03e00,
        3.03e00,
        3.14e00,
        3.15e00,
        3.16e00,
        3.17e00,
        3.17e00,
        3.18e00,
        3.18e00,
        3.19e00,
        3.20e00,
        3.22e00,
        3.24e00,
        3.25e00,
        3.29e00,
        3.31e00,
        3.32e00,
        3.32e00,
        3.34e00,
        3.35e00,
        3.36e00,
        3.38e00,
        3.44e00,
        3.45e00,
        3.46e00,
        3.48e00,
        3.49e00,
        3.53e00,
        3.59e00,
        3.63e00,
        3.70e00,
        3.70e00,
        3.76e00,
        3.80e00,
        3.80e00,
        3.80e00,
        3.83e00,
        3.84e00,
        3.88e00,
        3.90e00,
        3.91e00,
        3.96e00,
        3.97e00,
        3.97e00,
        4.02e00,
        4.03e00,
        4.06e00,
        4.12e00,
        4.19e00,
        4.21e00,
        4.53e00,
        4.56e00,
        4.61e00,
        4.62e00,
        4.73e00,
        5.13e00,
        5.21e00,
        5.40e00,
        5.98e00,
        6.12e00,
        6.94e00,
        7.38e00,
        7.56e00,
        8.06e00,
        1.38e01,
        1.51e01,
        1.82e01,
    ]
    return data


@pytest.fixture
@helpers.seed
def jp_data():
    pyplot.rcdefaults()
    N = 37
    df = pandas.DataFrame(
        {
            "A": numpy.random.normal(size=N),
            "B": numpy.random.lognormal(mean=0.25, sigma=1.25, size=N),
            "C": numpy.random.lognormal(mean=1.25, sigma=0.75, size=N),
        }
    )

    return df


@pytest.fixture
@helpers.seed
def cat_hist_data():
    N = 100
    years = [2011, 2012, 2013, 2014]
    df = pandas.DataFrame(
        {
            "depth": numpy.random.uniform(low=0.2, high=39, size=N),
            "year": numpy.random.choice(years, size=N),
            "has_outflow": numpy.random.choice([False, True], size=N),
        }
    )
    return df


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_rotateTickLabels_xaxis():
    fig, ax = pyplot.subplots()
    ax.set_xticks([1, 2, 3])
    ax.set_xticklabels(["AAA", "BBB", "CCC"])
    viz.rotate_tick_labels(ax, 60, "x")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_rotateTickLabels_yaxis():
    fig, ax = pyplot.subplots()
    ax.set_yticks([1, 2, 3])
    ax.set_yticklabels(["AAA", "BBB", "CCC"])
    viz.rotate_tick_labels(ax, -30, "y")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_rotateTickLabels_both():
    fig, ax = pyplot.subplots()
    ax.set_xticks([1, 2, 3])
    ax.set_xticklabels(["XXX", "YYY", "ZZZ"])

    ax.set_yticks([1, 2, 3])
    ax.set_yticklabels(["AAA", "BBB", "CCC"])
    viz.rotate_tick_labels(ax, 45, "both")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_log_formatter():
    fig, ax = pyplot.subplots()
    ax.plot([1, 5], [0.0005, 5e6])
    ax.set_yscale("log")
    ax.set_ylim([0.0001, 1e7])
    ax.yaxis.set_major_formatter(viz.log_formatter())
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_log_formatter_alt():
    fig, ax = pyplot.subplots()
    ax.plot([1, 5], [0.0005, 5e6])
    ax.set_yscale("log")
    ax.set_ylim([0.0001, 1e7])
    ax.yaxis.set_major_formatter(viz.log_formatter(use_1x=False))
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_log_formatter_alt_2():
    fig, (ax1, ax2) = pyplot.subplots(ncols=2)
    ax1.set_yscale("log")
    ax1.set_ylim((1e-4, 1e5))
    ax1.yaxis.set_major_formatter(viz.log_formatter(threshold=4))

    ax2.set_yscale("log")
    ax2.set_ylim((1e-7, 1e5))
    ax2.yaxis.set_major_formatter(viz.log_formatter(threshold=4, use_1x=False))
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_gridlines_basic(plot_data):
    fig, ax = pyplot.subplots()
    ax.plot(plot_data)
    viz.gridlines(ax, "xlabel", "ylabel")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_gridlines_ylog(plot_data):
    fig, ax = pyplot.subplots()
    ax.plot(plot_data)
    viz.gridlines(ax, "xlabel", "ylabel", yscale="log")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_gridlines_ylog_noyminor(plot_data):
    fig, ax = pyplot.subplots()
    ax.plot(plot_data)
    viz.gridlines(ax, "xlabel", "ylabel", yscale="log", yminor=False)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_one2one():
    fig, ax = pyplot.subplots()
    ax.set_xlim([-2, 5])
    ax.set_ylim([-3, 3])
    viz.one2one_line(ax, label="Equality", lw=5, ls="--")
    ax.legend()
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE * 1.6)
def test_jointplot_defaultlabels(jp_data):
    with seaborn.axes_style("ticks"):
        jg1 = viz.jointplot(x="B", y="C", data=jp_data, one2one=False, color="b")
    assert jg1.ax_joint.get_xlabel() == "B"
    assert jg1.ax_joint.get_ylabel() == "C"
    # nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_xlim()), [0, 16])
    # nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_ylim()), [0, 16])
    return jg1.figure


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE * 1.6)
def test_jointplot_xlabeled(jp_data):
    with seaborn.axes_style("ticks"):
        jg2 = viz.jointplot(
            x="B", y="C", data=jp_data, one2one=False, color="g", xlabel="Quantity B"
        )
    assert jg2.ax_joint.get_xlabel() == "Quantity B"
    return jg2.figure


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE * 1.6)
def test_jointplot_ylabeled(jp_data):
    with seaborn.axes_style("ticks"):
        jg3 = viz.jointplot(
            x="B", y="C", data=jp_data, one2one=False, color="r", ylabel="Quantity C"
        )
    assert jg3.ax_joint.get_ylabel() == "Quantity C"
    return jg3.figure


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE * 1.6)
def test_jointplot_bothlabeled(jp_data):
    with seaborn.axes_style("ticks"):
        jg4 = viz.jointplot(
            x="B",
            y="C",
            data=jp_data,
            one2one=False,
            color="k",
            xlabel="Quantity B",
            ylabel="Quantity C",
        )
    assert jg4.ax_joint.get_xlabel() == "Quantity B"
    assert jg4.ax_joint.get_ylabel() == "Quantity C"
    return jg4.figure


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE * 1.6)
def test_jointplot_zerominFalse(jp_data):
    with seaborn.axes_style("ticks"):
        jg1 = viz.jointplot(x="A", y="C", data=jp_data, zeromin=False, one2one=False)
    nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_xlim()), [-3, 3])
    nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_ylim()), [0, 16])
    return jg1.figure


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE * 1.6)
def test_jointplot_one2one(jp_data):
    with seaborn.axes_style("ticks"):
        jg1 = viz.jointplot(x="B", y="C", data=jp_data, one2one=True)
    nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_xlim()), [0, 16])
    nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_ylim()), [0, 16])
    return jg1.figure


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE * 1.6)
def test_jointplot_one2one_zerominFalse(jp_data):
    with seaborn.axes_style("ticks"):
        jg1 = viz.jointplot(x="A", y="C", data=jp_data, one2one=True, zeromin=False)
    nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_xlim()), [-3, 16])
    nptest.assert_array_equal(numpy.round(jg1.ax_joint.get_ylim()), [-3, 16])
    return jg1.figure


@pytest.mark.parametrize(
    ("case", "xform_in", "xform_out"),
    [
        ("linear", utils.no_op, utils.no_op),
        ("natlog", numpy.log, numpy.exp),
        ("log10", numpy.log10, lambda x: 10**x),
    ],
)
@pytest.mark.parametrize("key", ["whishi", "whislo", "fliers"])
def test_whiskers_and_fliers(whisk_flier_data, case, xform_in, xform_out, key):
    expected_results = {
        "linear": {
            "whishi": 4.62,
            "whislo": 0.00111,
            "fliers": numpy.array(
                [
                    4.730,
                    5.130,
                    5.210,
                    5.400,
                    5.980,
                    6.120,
                    6.940,
                    7.380,
                    7.560,
                    8.060,
                    13.800,
                    15.100,
                    18.200,
                ]
            ),
        },
        "natlog": {
            "whishi": 8.060,
            "whislo": 0.2700,
            "fliers": numpy.array(
                [
                    2.200e-01,
                    1.000e-03,
                    4.900e-02,
                    5.600e-02,
                    1.400e-01,
                    1.690e-01,
                    1.830e-01,
                    2.060e-01,
                    2.100e-01,
                    2.130e-01,
                    1.380e01,
                    1.510e01,
                    1.820e01,
                ]
            ),
        },
        "log10": {
            "whishi": 8.060,
            "whislo": 0.2700,
            "fliers": numpy.array(
                [
                    2.200e-01,
                    1.000e-03,
                    4.900e-02,
                    5.600e-02,
                    1.400e-01,
                    1.690e-01,
                    1.830e-01,
                    2.060e-01,
                    2.100e-01,
                    2.130e-01,
                    1.380e01,
                    1.510e01,
                    1.820e01,
                ]
            ),
        },
    }
    q1 = numpy.percentile(whisk_flier_data, 25)
    q3 = numpy.percentile(whisk_flier_data, 75)
    result = viz.whiskers_and_fliers(
        xform_in(whisk_flier_data), xform_in(q1), xform_in(q3), transformout=xform_out
    )

    expected = expected_results[case]
    nptest.assert_array_almost_equal(result[key], expected[key], decimal=3)


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_boxplot_basic(boxplot_data):
    fig, ax = pyplot.subplots()
    viz.boxplot(boxplot_data, ax=ax, shownotches=True, patch_artist=False)
    ax.set_xlim((0, 2))
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_boxplot_with_mean(boxplot_data):
    fig, ax = pyplot.subplots()
    viz.boxplot(
        boxplot_data,
        ax=ax,
        shownotches=True,
        patch_artist=True,
        marker="^",
        color="red",
        showmean=True,
    )
    ax.set_xlim((0, 2))
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_probplot_prob(plot_data):
    fig, ax = pyplot.subplots()
    fig = viz.probplot(plot_data, ax=ax, xlabel="Test xlabel")
    assert isinstance(fig, pyplot.Figure)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_probplot_qq(plot_data):
    fig, ax = pyplot.subplots()
    fig = viz.probplot(
        plot_data, ax=ax, axtype="qq", ylabel="Test label", scatter_kws=dict(color="r")
    )
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_probplot_pp(plot_data):
    fig, ax = pyplot.subplots()
    scatter_kws = dict(color="b", linestyle="--", markeredgecolor="g", markerfacecolor="none")
    fig = viz.probplot(
        plot_data,
        ax=ax,
        axtype="pp",
        yscale="linear",
        xlabel="test x",
        ylabel="test y",
        scatter_kws=scatter_kws,
    )
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_probplot_prob_bestfit(plot_data):
    fig, ax = pyplot.subplots()
    fig = viz.probplot(plot_data, ax=ax, xlabel="Test xlabel", bestfit=True)
    assert isinstance(fig, pyplot.Figure)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_probplot_qq_bestfit(plot_data):
    fig, ax = pyplot.subplots()
    fig = viz.probplot(plot_data, ax=ax, axtype="qq", bestfit=True, ylabel="Test label")
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_probplot_pp_bestfit(plot_data):
    fig, ax = pyplot.subplots()
    scatter_kws = {"marker": "s", "color": "red"}
    line_kws = {"linestyle": "--", "linewidth": 3}
    fig = viz.probplot(
        plot_data,
        ax=ax,
        axtype="pp",
        yscale="linear",
        xlabel="test x",
        bestfit=True,
        ylabel="test y",
        scatter_kws=scatter_kws,
        line_kws=line_kws,
    )
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test__connect_spines():
    fig, ((ax1, ax2), (ax3, ax4)) = pyplot.subplots(ncols=2, nrows=2)
    viz._connect_spines(ax1, ax2, 0.5, 0.75)
    viz._connect_spines(ax3, ax4, 0.9, 0.1, linewidth=2, color="r", linestyle="dashed")
    return fig


@pytest.mark.xfail
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_parallel_coordinates():
    df = seaborn.load_dataset("iris")
    fig = viz.parallel_coordinates(df, hue="species")
    return fig


@pytest.mark.xfail(TOLERANCE > 15, reason="GH Action weirdness")
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_categorical_histogram_simple(cat_hist_data):
    with seaborn.axes_style("ticks"):
        bins = numpy.arange(5, 35, 5)
        fig = viz.categorical_histogram(cat_hist_data, "depth", bins)
        return fig.figure


@pytest.mark.xfail(TOLERANCE > 15, reason="GH Action weirdness")
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_categorical_histogram_complex(cat_hist_data):
    with seaborn.axes_style("ticks"):
        bins = numpy.arange(5, 35, 5)
        fig = viz.categorical_histogram(cat_hist_data, "depth", bins, hue="year", row="has_outflow")
        return fig.figure



from functools import partial
from io import StringIO
from textwrap import dedent

import numpy
import pandas
import pandas.testing as pdtest
import pytest

from wqio.tests import helpers
from wqio.utils import misc


@pytest.fixture
def basic_data():
    testcsv = """\
    Date,A,B,C,D
    X,1,2,3,4
    Y,5,6,7,8
    Z,9,0,1,2
    """
    return pandas.read_csv(StringIO(dedent(testcsv)), index_col=["Date"])


@pytest.fixture
def multiindex_df():
    index = pandas.MultiIndex.from_product([["A", "B", "C"], ["mg/L"]], names=["loc", "units"])
    return pandas.DataFrame([[1, 2], [3, 4], [5, 6]], index=index, columns=["a", "b"])


class mockDataset:
    def __init__(self, inflow, outflow):
        self.inflow = mockLocation(inflow)
        self.outflow = mockLocation(outflow)


class mockLocation:
    def __init__(self, data):
        self.data = data
        self.stats = mockSummary(data)


class mockSummary:
    def __init__(self, data):
        self.N = len(data)
        self.max = max(data)
        self.min = min(data)
        self.nonething = None


def test_add_column_level(basic_data):
    known_cols = pandas.MultiIndex.from_tuples(
        [("test", "A"), ("test", "B"), ("test", "C"), ("test", "D")]
    )
    newdata = misc.add_column_level(basic_data, "test", "testlevel")
    assert known_cols.tolist() == newdata.columns.tolist()

    # can only add levels to non-MultiIndex columns
    with helpers.raises(ValueError):
        misc.add_column_level(newdata, "test2", "testlevel2")


@pytest.mark.parametrize("L1", [0, "loc"])
@pytest.mark.parametrize("L2", [2, "units"])
def test_swap_column_levels(multiindex_df, L1, L2):
    columns = pandas.MultiIndex.from_product(
        [["A", "B", "C"], ["res", "cen"], ["mg/L"]], names=["loc", "value", "units"]
    )
    data = numpy.arange(len(columns) * 10).reshape((10, len(columns)))
    df = pandas.DataFrame(data, columns=columns).pipe(misc.swap_column_levels, L1, L2)

    expected_columns = pandas.MultiIndex.from_product(
        [["mg/L"], ["cen", "res"], ["A", "B", "C"]], names=["units", "value", "loc"]
    )

    pdtest.assert_index_equal(df.columns, expected_columns)


def test_flatten_columns(multiindex_df, basic_data):
    expected = ["A_mg/L", "B_mg/L", "C_mg/L"]
    flat = misc.flatten_columns(multiindex_df.T)
    assert flat.columns.tolist() == expected
    assert misc.flatten_columns(basic_data).columns.tolist() == basic_data.columns.tolist()


def test_expand_columns():
    x = numpy.arange(12).reshape(3, 4)
    df = pandas.DataFrame(x, columns=("A_a", "A_b", "B_a", "B_c"))

    res_cols = pandas.MultiIndex(
        levels=[["A", "B"], ["a", "b", "c"]],
        codes=[[0, 0, 1, 1], [0, 1, 0, 2]],
        names=["top", "bottom"],
    )
    expected = pandas.DataFrame(x, columns=res_cols)
    result = misc.expand_columns(df, ["top", "bottom"])
    pdtest.assert_frame_equal(result, expected)


@pytest.mark.parametrize("criteria", [None, lambda row: row[0] in ["A", "B"]])
@pytest.mark.parametrize("dropold", [True, False])
def test_redefine_index_level(multiindex_df, criteria, dropold):
    expected_cols = ["a", "b"]
    if dropold:
        expected_value = [[1, 2], [3, 4], [5, 6]]
        if criteria:
            expected_index = [("A", "ug/L"), ("B", "ug/L"), ("C", "mg/L")]
        else:
            expected_index = [("A", "ug/L"), ("B", "ug/L"), ("C", "ug/L")]

    else:
        if criteria:
            expected_value = [[1, 2], [1, 2], [3, 4], [3, 4], [5, 6]]
            expected_index = [
                ("A", "mg/L"),
                ("A", "ug/L"),
                ("B", "mg/L"),
                ("B", "ug/L"),
                ("C", "mg/L"),
            ]
        else:
            expected_value = [[1, 2], [1, 2], [3, 4], [3, 4], [5, 6], [5, 6]]
            expected_index = [
                ("A", "mg/L"),
                ("A", "ug/L"),
                ("B", "mg/L"),
                ("B", "ug/L"),
                ("C", "mg/L"),
                ("C", "ug/L"),
            ]

    result = misc.redefine_index_level(
        multiindex_df, "units", "ug/L", criteria=criteria, dropold=dropold
    )
    expected = pandas.DataFrame(
        data=expected_value,
        index=pandas.MultiIndex.from_tuples(expected_index, names=["loc", "units"]),
        columns=expected_cols,
    )
    pdtest.assert_frame_equal(result, expected)


@pytest.fixture
def basic_dataset():
    inflow = [1, 3, 4, 12.57]
    outflow = [2, 5, 7, 15.17]
    return mockDataset(inflow, outflow)


def test_nested_getattr(basic_dataset):
    result = misc.nested_getattr(basic_dataset, "inflow.stats.max")
    expected = basic_dataset.inflow.stats.max
    assert result == expected


@pytest.mark.parametrize(
    ("strformat", "expected", "attribute"),
    [
        ("%d", "4", "inflow.stats.N"),
        ("%0.2f", "15.17", "outflow.stats.max"),
        (None, "--", "inflow.stats.nonething"),
    ],
)
def test_stringify(basic_dataset, strformat, expected, attribute):
    result = misc.stringify(basic_dataset, strformat, attribute=attribute)
    assert result == expected


def test_categorize_columns():
    csvdata = StringIO(
        dedent(
            """\
        parameter,units,season,lower,NSQD Median,upper
        Cadmium (Cd),ug/L,autumn,0.117,0.361,0.52
        Cadmium (Cd),ug/L,spring,0.172,0.352,0.53
        Cadmium (Cd),ug/L,summer,0.304,0.411,0.476
        Cadmium (Cd),ug/L,winter,0.355,0.559,1.125
        Dissolved Chloride (Cl),mg/L,autumn,0.342,2.3,5.8
        Dissolved Chloride (Cl),mg/L,spring,2.5,2.5,2.5
        Dissolved Chloride (Cl),mg/L,summer,0.308,0.762,1.24
        Escherichia coli,MPN/100 mL,autumn,1200.0,15500.0,24000.0
        Escherichia coli,MPN/100 mL,spring,10.0,630.0,810.0
        Escherichia coli,MPN/100 mL,summer,21000.0,27000.0,35000.0
        Escherichia coli,MPN/100 mL,winter,20.0,200.0,800.0
    """
        )
    )
    df = pandas.read_csv(csvdata)
    df2 = misc.categorize_columns(df, "parameter", "units", "season")

    # check pandas API that the first df has object columns
    assert object in df.dtypes.values

    # confirm that all of those objects are gone
    assert object not in df2.dtypes.values

    with helpers.raises(ValueError):
        misc.categorize_columns(df, "parameter", "upper")


@pytest.mark.parametrize(
    ("value", "expected"), [(3, "<5"), (17, "15 - 20"), (25, "20 - 25"), (46, ">35")]
)
@pytest.mark.parametrize("units", [None, "mm"])
def test_classifier(value, units, expected):
    bins = numpy.arange(5, 36, 5)
    if units is not None:
        expected = f"{expected} {units}"

    result = misc.classifier(value, bins, units=units)
    assert result == expected
    assert numpy.isnan(misc.classifier(numpy.nan, bins, units=units))


def test_unique_categories():
    bins = [5, 10, 15]
    classifier = partial(misc.classifier, bins=bins, units="mm")

    known_categories = ["<5 mm", "5 - 10 mm", "10 - 15 mm", ">15 mm"]
    result_categories = misc.unique_categories(classifier, bins)

    assert result_categories == known_categories


def test_pop_many():
    some_dict = dict(zip(list("ABCDE"), range(5)))
    expected = {"C": 2, "D": 3}
    assert misc.pop_many(some_dict, "A", "B", "E") == expected


def test_selector():
    x = numpy.arange(10)
    expected = numpy.array(list("AAABBBCCZZ"))
    result = misc.selector("Z", (x <= 2, "A"), (x < 6, "B"), (x <= 7, "C"))
    assert all(result == expected)


@pytest.mark.parametrize("input_values", ["A", 4, ("this", "5")])
def test_non_filter(input_values):
    assert misc.non_filter(input_values)


@pytest.mark.parametrize("input_values", ["A", 4, ("this", "5")])
def test_no_op(input_values):
    assert misc.no_op(input_values) == input_values


@pytest.mark.parametrize("value", [4, lambda x: 4])
def test_assign_multilevel_column(value):
    df = pandas.DataFrame(
        data=1,
        index=pandas.MultiIndex.from_product([list("ABCD"), [1, 2, 3, 4]]),
        columns=pandas.MultiIndex.from_product([list("abc"), [1, 2, 3]]),
    )

    result = misc.assign_multilevel_column(df, value, "d", 1)
    expected = pandas.Series(4, index=df.index, name=("d", 1))
    pdtest.assert_series_equal(result[("d", 1)], expected)


@pytest.mark.parametrize("join_char", [None, "-"])
def test_symbolize_bools(join_char):
    df = pandas.DataFrame(
        {
            "A": [True, False, False],
            "B": [False, True, True],
            "C": [False, True, numpy.nan],
        }
    )

    result = misc.symbolize_bools(
        df, true_symbol="◆", false_symbol="◇", other_symbol="✖", join_char=join_char
    )
    if not join_char:
        expected = pandas.DataFrame(
            {
                "A": {0: "◆", 1: "◇", 2: "◇"},
                "B": {0: "◇", 1: "◆", 2: "◆"},
                "C": {0: "◇", 1: "◆", 2: "✖"},
            }
        )
        pdtest.assert_frame_equal(result, expected)
    else:
        expected = pandas.Series(["◆-◇-◇", "◇-◆-◆", "◇-◆-✖"])
        pdtest.assert_series_equal(result, expected)



import logging
import warnings

import numpy
import pandas
from scipy import stats

from wqio import utils

_logger = logging.getLogger(__name__)


def _ros_sort(df, result, censorship, log=True, warn=False):
    """
    This function prepares a dataframe for ROS. It sorts ascending with
    left-censored observations on top. Censored results larger than the
    maximum uncensored results are removed from the dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
    result : str
        Name of the column in the dataframe that contains observed
        values. Censored values should be set to the detection (upper)
        limit.
    censorship : str
        Name of the column in the dataframe that indicates that a
        result is left-censored. (i.e., True -> censored,
        False -> uncensored)

    Returns
    ------
    sorted_df : pandas.DataFrame
        The sorted dataframe with all columns dropped except the
        result and censorship columns.

    """

    # separate uncensored data from censored data
    max_uncensored = df.loc[~df[censorship], result].max()
    if (df.loc[df[censorship], result] > max_uncensored).any():
        msg = "Dropping censored results greater than the max uncensored result."
        utils.log_or_warn(
            msg,
            warning=UserWarning if warn else None,
            logger=_logger.debug if log else None,
        )

    df_sorted = (
        df[[censorship, result]]
        .sort_values(by=[censorship, result], ascending=[False, True])
        .where(
            lambda df: (~df[censorship])
            | (  # uncensored values
                (df[result] <= max_uncensored) & df[censorship]
            )  # censored values < max_uncen
        )
        .dropna(how="all")
        .reset_index(drop=True)
        .assign(**{censorship: lambda df: df[censorship].astype(bool)})
    )
    return df_sorted[[result, censorship]]


def cohn_numbers(df, result, censorship):
    r"""
    Computes the Cohn numbers for the detection limits in the dataset.

    The Cohn Numbers are:

        - :math:`A_j =` the number of uncensored obs above the
          :math:`j^\mathrm{th}` threshold.
        - :math:`B_j =` the number of observations (cen & uncen) below
          the :math:`j^\mathrm{th}` threshold.
        - :math:`C_j =` the number of censored observations at the
          :math:`j^\mathrm{th}` threshold.
        - :math:`\mathrm{PE}_j =` the probability of exceeding the
          :math:`j^\mathrm{th}` threshold
        - :math:`\mathrm{DL}_j =` the unique, sorted detection limits
        - :math:`\mathrm{DL}_{j+1} = \mathrm{DL}_j` shifted down a
          single index (row)

    Parameters
    ----------
    dataframe : pandas.DataFrame
    result : str
        Name of the column in the dataframe that contains observed
        values. Censored values should be set to the detection (upper)
        limit.
    censorship : str
        Name of the column in the dataframe that indicates that a
        result is left-censored. (i.e., True -> censored,
        False -> uncensored)

    Returns
    -------
    cohn : pandas.DataFrame

    """

    def nuncen_above(row):
        """A, the number of uncensored obs above the given threshold."""

        # index of results above the lower_dl DL
        above = df[result] >= row["lower_dl"]

        # index of results below the upper_dl DL
        below = df[result] < row["upper_dl"]

        # index of non-detect results
        detect = df[censorship].eq(False)

        # return the number of results where all conditions are True
        return df[above & below & detect].shape[0]

    def nobs_below(row):
        """B, the number of observations (cen & uncen) below the given
        threshold
        """

        # index of data less than the lower_dl DL
        less_than = df[result] < row["lower_dl"]

        # index of data less than or equal to the lower_dl DL
        less_thanequal = df[result] <= row["lower_dl"]

        # index of detects, non-detects
        uncensored = df[censorship].eq(False)
        censored = df[censorship].eq(True)

        # number results less than or equal to lower_dl DL and non-detect
        LTE_censored = df[less_thanequal & censored].shape[0]

        # number of results less than lower_dl DL and detected
        LT_uncensored = df[less_than & uncensored].shape[0]

        # return the sum
        return LTE_censored + LT_uncensored

    def ncen_equal(row):
        """C, the number of censored observations at the given
        threshold.
        """

        censored_index = df[censorship]
        censored_data = df[result][censored_index]
        censored_below = censored_data == row["lower_dl"]
        return censored_below.sum()

    def set_upper_limit(cohn):
        """Sets the upper_dl DL for each row of the Cohn dataframe."""
        if cohn.shape[0] > 1:
            return cohn["lower_dl"].shift(-1).fillna(value=numpy.inf)
        else:
            return [numpy.inf]

    def compute_PE(A, B):
        """Computes the probability of excedance for each row of the
        Cohn dataframe."""
        N = len(A)
        PE = numpy.empty(N, dtype="float64")
        PE[-1] = 0.0
        for j in range(N - 2, -1, -1):
            PE[j] = PE[j + 1] + (1 - PE[j + 1]) * A[j] / (A[j] + B[j])

        return PE

    # unique, sorted detection limts
    censored_data = df[censorship]
    DLs = pandas.unique(df.loc[censored_data, result])
    DLs.sort()

    # if there is a results smaller than the minimum detection limit,
    # add that value to the array
    if DLs.shape[0] > 0:
        if df[result].min() < DLs.min():
            DLs = numpy.hstack([df[result].min(), DLs])

        # create a dataframe
        cohn = (
            pandas.DataFrame(DLs, columns=["lower_dl"])
            .assign(upper_dl=lambda df: set_upper_limit(df))
            .assign(nuncen_above=lambda df: df.apply(nuncen_above, axis=1))
            .assign(nobs_below=lambda df: df.apply(nobs_below, axis=1))
            .assign(ncen_equal=lambda df: df.apply(ncen_equal, axis=1))
            .reindex(range(DLs.shape[0] + 1))
            .assign(prob_exceedance=lambda df: compute_PE(df["nuncen_above"], df["nobs_below"]))
        )

    else:
        dl_cols = [
            "lower_dl",
            "upper_dl",
            "nuncen_above",
            "nobs_below",
            "ncen_equal",
            "prob_exceedance",
        ]
        cohn = pandas.DataFrame(numpy.empty((0, len(dl_cols))), columns=dl_cols)

    return cohn


def _detection_limit_index(res, cohn):
    """Helper function to create an array of indices for the detection
    limits (cohn) corresponding to each data point.

    Parameters
    ----------
    res : float
        A single observed result from the larger dataset.
    cohn : pandas.DataFrame
        Dataframe of Cohn numbers.

    Returns
    -------
    det_limit_index : int
        The index of the corresponding detection limit in `cohn`

    See also
    --------
    cohn_numbers

    """

    if cohn.shape[0] > 0:
        (index,) = numpy.where(cohn["lower_dl"] <= res)
        det_limit_index = index[-1]
    else:
        det_limit_index = 0

    return det_limit_index


def _ros_group_rank(df, dl_idx, censorship):
    """
    Ranks each result within the groups defined by the record's
    detection limit index and censorship.

    Parameters
    ----------
    df : pandas.DataFrame
    dl_idx : str
        Name of the column in the dataframe the index of the result's
        corresponding detection limit in the `cohn` dataframe.
    censorship : str
        Name of the column in the dataframe that indicates that a
        result is left-censored. (i.e., True -> censored,
        False -> uncensored)

    Returns
    -------
    ranks : numpy.array
        Array of ranks for the dataset.

    """

    ranks = (
        df.assign(rank=1).groupby(by=[dl_idx, censorship])["rank"].transform(lambda g: g.cumsum())
    )
    return ranks


def _ros_plot_pos(row, censorship, cohn):
    """
    Compute the ROS plotting position for a result based on its rank,
    censorship, detection limit index.

    Parameters
    ----------
    row : pandas.Series or dict-like
        Full observation (row) from a censored dataset. Requires a
        'rank', 'detection_limit', and `censorship` column.
    censorship : str
        Name of the column in the dataframe that indicates that a
        result is left-censored. (i.e., True -> censored,
        False -> uncensored)
    cohn : pandas.DataFrame
        Dataframe of Cohn numbers.

    Returns
    -------
    plotting_position : float

    See also
    --------
    cohn_numbers

    """

    DL_index = row["det_limit_index"]
    rank = row["rank"]
    censored = row[censorship]

    dl_1 = cohn.iloc[DL_index]
    dl_2 = cohn.iloc[DL_index + 1]
    if censored:
        return (1 - dl_1["prob_exceedance"]) * rank / (dl_1["ncen_equal"] + 1)
    else:
        return (1 - dl_1["prob_exceedance"]) + (
            dl_1["prob_exceedance"] - dl_2["prob_exceedance"]
        ) * rank / (dl_1["nuncen_above"] + 1)


def _norm_plot_pos(results):
    """
    Computes standard normal (Gaussian) plotting positions using scipy.

    Parameters
    ----------
    results : array-like
        Sequence of observed quantities.

    Returns
    -------
    plotting_position : array of floats

    """
    ppos, sorted_res = stats.probplot(results, fit=False)
    return stats.norm.cdf(ppos)


def plotting_positions(df, censorship, cohn):
    """
    Compute the ROS plotting positions for results based on their rank,
    censorship, detection limit index.

    Parameters
    ----------
    df : pandas.DataFrame.
    censorship : str
        Name of the column in the dataframe that indicates that a
        result is left-censored. (i.e., True -> censored,
        False -> uncensored)
    cohn : pandas.DataFrame
        Dataframe of Cohn numbers.

    Returns
    -------
    plotting_position : array of float

    See also
    --------
    cohn_numbers

    """

    plot_pos = df.apply(lambda r: _ros_plot_pos(r, censorship, cohn), axis=1)

    # correctly sort the plotting positions of the ND data:
    ND_plotpos = plot_pos[df[censorship]]
    ND_plotpos.values.sort()
    plot_pos.loc[df[censorship]] = ND_plotpos

    return plot_pos


def _ros_estimate(df, result, censorship, transform_in, transform_out):
    """Computed the estimated censored from the best-fit line of a
    probability plot of the uncensored values.

    Parameters
    ----------
    df : pandas.DataFrame
    result : str
        Name of the column in the dataframe that contains observed
        values. Censored values should be set to the detection (upper)
        limit.
    censorship : str
        Name of the column in the dataframe that indicates that a
        result is left-censored. (i.e., True -> censored,
        False -> uncensored)
    transform_in, transform_out : callable
        Transformations to be applied to the data prior to fitting
        the line and after estimated values from that line. Typically,
        `numpy.log` and `numpy.exp` are used, respectively.

    Returns
    -------
    estimated : pandas.DataFrame
        A new dataframe with two new columns: "estimated" and "final".
        The "estimated" column contains of the values inferred from the
        best-fit line. The "final" column contains the estimated values
        only where the original results were censored, and the original
        results everwhere else.

    """

    # detect/non-detect selectors
    uncensored_mask = df[censorship].eq(False)
    censored_mask = df[censorship].eq(True)

    # fit a line to the logs of the detected data
    fit_params = stats.linregress(
        df["Zprelim"][uncensored_mask], transform_in(df[result][uncensored_mask])
    )

    # pull out the slope and intercept for use later
    slope, intercept = fit_params[:2]

    # model the data based on the best-fit curve
    df = df.assign(
        estimated=transform_out(slope * df["Zprelim"][censored_mask] + intercept)
    ).assign(final=lambda df: numpy.where(df[censorship], df["estimated"], df[result]))

    return df


def _do_ros(
    df,
    result,
    censorship,
    transform_in,
    transform_out,
    floor=None,
    log=True,
    warn=False,
):
    """
    Prepares a dataframe for, and then esimates the values of a censored
    dataset using Regression on Order Statistics

    Parameters
    ----------
    df : pandas.DataFrame
    result : str
        Name of the column in the dataframe that contains observed
        values. Censored values should be set to the detection (upper)
        limit.
    censorship : str
        Name of the column in the dataframe that indicates that a
        result is left-censored. (i.e., True -> censored,
        False -> uncensored)
    transform_in, transform_out : callable
        Transformations to be applied to the data prior to fitting
        the line and after estimated values from that line. Typically,
        `numpy.log` and `numpy.exp` are used, respectively.
    floor : float (optional)
        When provided, all resulting values (inputed or otherwise) that are less
        than the `floor` will be replaced with its value. As an example, this is
        useful with bacterial concentrations since a value less than 1 isn't
        feasible.

    Returns
    -------
    estimated : pandas.DataFrame
        A new dataframe with two new columns: "estimated" and "final".
        The "estimated" column contains of the values inferred from the
        best-fit line. The "final" column contains the estimated values
        only where the original results were censored, and the original
        results everwhere else.

    """

    # compute the Cohn numbers
    cohn = cohn_numbers(df, result=result, censorship=censorship)
    with warnings.catch_warnings():
        warnings.simplefilter("once")
        modeled = (
            df.pipe(_ros_sort, result=result, censorship=censorship, log=log, warn=warn)
            .assign(
                det_limit_index=lambda df: df[result].apply(_detection_limit_index, args=(cohn,))
            )
            .assign(rank=lambda df: _ros_group_rank(df, "det_limit_index", censorship))
            .assign(plot_pos=lambda df: plotting_positions(df, censorship, cohn))
            .assign(Zprelim=lambda df: stats.norm.ppf(df["plot_pos"]))
            .pipe(_ros_estimate, result, censorship, transform_in, transform_out)
        )

    if floor:
        modeled = modeled.assign(final=modeled["final"].where(lambda x: x >= floor, floor))

    return modeled


def is_valid_to_ros(df, censorship, max_fraction_censored=0.8, min_uncensored=2, as_obj=False):
    # basic counts/metrics of the dataset
    N_observations = df.shape[0]
    N_censored = df[censorship].astype(int).sum()
    N_uncensored = N_observations - N_censored
    fraction_censored = N_censored / N_observations

    enough_uncensored = N_uncensored >= min_uncensored
    not_too_many_censored = fraction_censored <= max_fraction_censored

    if as_obj:
        return {
            "enough_uncensored": enough_uncensored,
            "not_too_many_censored": not_too_many_censored,
        }
    return enough_uncensored and not_too_many_censored


def ROS(
    result,
    censorship,
    df=None,
    min_uncensored=2,
    max_fraction_censored=0.8,
    substitution_fraction=0.5,
    transform_in=numpy.log,
    transform_out=numpy.exp,
    as_array=True,
    floor=None,
    log=True,
    warn=False,
):
    """
    Impute censored dataset using Regression on Order Statistics (ROS)
    or simple substitution if insufficient uncensored data exists.

    Method described in *Nondetects and Data Analysis* by Dennis R.
    Helsel (John Wiley, 2005) to estimate the left-censored (non-detect)
    values of a dataset.

    Parameters
    ----------
    result : str or array-like
        Label of the column or the float array of censored results

    censorship : str
        Label of the column or the bool array of the censorship
        status of the results.

          * True if censored,
          * False if uncensored

    df : pandas.DataFrame, optional
        If `result` and `censorship` are labels, this is the DataFrame
        that contains those columns.

    min_uncensored : int (default is 2)
        The minimum number of uncensored values required before ROS
        can be used to impute the censored results. When this criterion
        is not met, simple substituion is used instead.

    max_fraction_censored : float (default is 0.8)
        The maximum fraction of censored data below which ROS can be
        used to impute the censored results. When this fraction is
        exceeded, simple substituion is used instead.

    substitution_fraction : float (default is 0.5)
        The fraction of the detection limit to be used during simple
        substitution of the censored values.

    transform_in : callable (default is numpy.log)
        Transformation to be applied to the values prior to fitting a
        line to the plotting positions vs. uncensored values.

    transform_out : callable (default is numpy.exp)
        Transformation to be applied to the imputed censored values
        estimated from the previously computed best-fit line.

    as_array : bool (default is True)
        When True, a numpy array of the imputed results is returned.
        Otherwise, a modified copy of the original dataframe with all
        of the intermediate calculations is returned.

    floor : float (optional)
        When provided, all resulting values (inputed or otherwise) that are less
        than the `floor` will be replaced with its value. As an example, this is
        useful with bacterial concentrations since a value less than 1 isn't
        feasible.

    Returns
    -------
    imputed : numpy.array (default) or pandas.DataFrame
        The final results where the censored values have either been
        imputed through ROS or substituted as a fraction of the
        detection limit.

    """

    # process arrays into a dataframe, if necessary
    if df is None:
        df = pandas.DataFrame({"res": result, "cen": censorship})
        result = "res"
        censorship = "cen"

    # add plotting positions if there are no censored values
    if df[censorship].astype(int).sum() == 0:
        output = df[[result, censorship]].assign(final=df[result])

    # normal ROS stuff
    elif is_valid_to_ros(
        df,
        censorship,
        max_fraction_censored=max_fraction_censored,
        min_uncensored=min_uncensored,
        as_obj=False,
    ):
        output = _do_ros(
            df,
            result,
            censorship,
            transform_in,
            transform_out,
            floor=floor,
            log=log,
            warn=warn,
        )

    # substitute w/ fraction of the DLs if there's insufficient
    # uncensored data
    else:
        final = numpy.where(df[censorship], df[result] * substitution_fraction, df[result])
        output = df.assign(final=final)[[result, censorship, "final"]]

    # convert to an array if necessary
    if as_array:
        output = output["final"].values

    return output



import warnings
from copy import copy
from functools import wraps

import numpy
import pandas


def head_tail(df, N=5):
    return pandas.concat([df.head(N), df.tail(N)])


def add_column_level(df, levelvalue, levelname):
    """Adds a second level to the column-index if a dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        The original dataframe to be modified.
    levelvalue : int or string
        Constant value to be assigned to the second level.
    levelname : string
        The name of the second level.

    Returns
    -------
    newdf : pandas.DataFrame
        The mutated dataframe with a MultiIndex in the columns.

    Example
    -------
    >>> df = pandas.DataFrame(columns=['res', 'qual'], index=range(3))
    >>> df.columns
    Index(['res', 'qual'], dtype='object')
    >>> df2 = add_column_level(df, 'Infl', 'location')
    >>> df2.columns
    MultiIndex([('Infl',  'res'),
                ('Infl', 'qual')],
               names=['location', 'quantity'])

    """

    if isinstance(df.columns, pandas.MultiIndex):
        raise ValueError("Dataframe already has MultiIndex on columns")

    origlevel = df.columns.names[0] or "quantity"
    return df.add_prefix(levelvalue + "_____").pipe(
        expand_columns, [levelname, origlevel], sep="_____"
    )


def swap_column_levels(df, level_1, level_2, sort=True):
    """Swaps columns levels in a dataframe with multi-level columns

    Parameters
    ----------
    df : pandas.DataFrame
        A dataframe with a multi-level column index
    level_1, level_2 : int or string
        The position or label of the columns levels to be swapped.

    Returns
    -------
    swapped : pandas.DataFrame

    Examples
    --------
    >>> import numpy
    >>> import pandas
    >>> import wqio
    >>> columns = pandas.MultiIndex.from_product(
    ...     [['A', 'B',], ['res', 'cen'], ['mg/L']],
    ...     names=['loc', 'value', 'units']
    ... )
    >>> data = numpy.arange(len(columns) * 3).reshape((3, len(columns)))
    >>> df = pandas.DataFrame(data, columns=columns)
    >>> df.columns
    MultiIndex([('A', 'res', 'mg/L'),
                ('A', 'cen', 'mg/L'),
                ('B', 'res', 'mg/L'),
                ('B', 'cen', 'mg/L')],
               names=['loc', 'value', 'units'])


    >>> swapped = wqio.utils.swap_column_levels(df, 'units', 'loc')
    >>> swapped.columns
    MultiIndex([('mg/L', 'cen', 'A'),
                ('mg/L', 'cen', 'B'),
                ('mg/L', 'res', 'A'),
                ('mg/L', 'res', 'B')],
               names=['units', 'value', 'loc'])


    """

    df2 = df.swaplevel(level_1, level_2, axis="columns")
    if sort:
        df2 = df2.sort_index(axis="columns")

    return df2


def flatten_columns(df: pandas.DataFrame, sep: str = "_"):
    """Completely flattens a multi-level column index

    Parameters
    ----------
    df : pandas.DataFrame
        The DataFrame with multi-level columns that will be flattened.
    sep : string, optional
        The string that will be used to delimit each level of the
        column index in the flattened names.

    Returns
    -------
    flattened : pandas.DataFrame

    """
    newcols = [sep.join(_) for _ in df.columns]
    return df.set_axis(newcols, axis="columns")


def expand_columns(df, names, sep="_"):
    """
    Expands a dataframe's columns into a multi-level index

    Parameters
    ----------
    df : pandas.DataFrame
    names : list of string
        The names for the new column levels.
    sep : str (default = '_')
        Character(s) on which the column labels will be split

    Returns
    -------
    pandas.DataFrame

    Example
    -------
    >>> import pandas
    >>> import numpy
    >>> from wqio import utils
    >>> x = numpy.arange(12).reshape(3, 4)
    >>> df = pandas.DataFrame(x, columns=('A_a','A_b', 'B_a', 'B_c'))
    >>> expanded = utils.expand_columns(df, ['top', 'bottom'], sep='_')
    >>> expanded.columns
    MultiIndex([('A', 'a'),
                ('A', 'b'),
                ('B', 'a'),
                ('B', 'c')],
               names=['top', 'bottom'])


    """

    newcols = df.columns.str.split(sep, expand=True)
    return df.set_axis(newcols, axis="columns").rename_axis(names, axis="columns")


def redefine_index_level(df, levelname, value, criteria=None, dropold=True):
    """Redefine a index values in a dataframe.

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe to be modified.
    levelname : string
        The name of the index level that needs to be modified. The catch
        here is that this value needs to be valid after calling
        `df.reset_index()`. In otherwords, if you have a 3-level
        column index and you want to modify the "Units" level of the
        index, you should actually pass `("Units", "", "")`. Annoying,
        but that's life right now.
    value : string or int
        The replacement value for the index level.
    criteria : function/lambda expression or None
        This should return True/False in a manner consitent with the
        `.select()` method of a pandas dataframe. See that docstring
        for more info. If None, the redifinition will apply to the whole
        df.
    dropold : optional bool (defaul is True)
        Toggles the replacement (True) or addition (False) of the data
        of the redefined BMPs into the the `data` df.

    Returns
    -------
    appended : pandas.DataFrame
        Dataframe with the modified index.

    """

    if criteria is None:

        def criteria(*args, **kwargs):
            return True

    redefined = (
        df.loc[df.index.map(criteria), :]
        .reset_index()
        .assign(**{levelname: value})
        .set_index(df.index.names)
    )

    if dropold:
        df = df.loc[df.index.map(lambda r: not criteria(r)), :]

    return pandas.concat([df, redefined]).sort_index()


def categorize_columns(df, *columns):
    newdf = df.copy()
    for c in columns:
        if newdf[c].dtype != object:
            raise ValueError(f"column {c} is not an object type")
        newdf[c] = newdf[c].astype("category")

    return newdf


def nested_getattr(baseobject, attribute):
    """Returns the value of an attribute of an object that is nested
    several layers deep.

    Parameters
    ----------
    baseobject : this seriously can be anything
        The top-level object
    attribute : string
        Any string of what you want.

    Returns
    -------
    output : object
        No telling what type it is. It depends on what you ask for.

    Examples
    --------
    >>> class Level_1:
    ...     a = 1
    >>> class Level_2:
    ...     b = Level_1()
    >>> x = Level_2()
    >>> nested_getattr(x, 'b.a')
    1

    """

    for attr in attribute.split("."):
        baseobject = getattr(baseobject, attr)
    return baseobject


def stringify(value, fmt, attribute=None):
    """Weird wrapper to format attributes of objects as strings

    Parameters
    ----------
    value : object
        The item or top-level container of the item to be formatted.
    fmt : str
        A valid old-style(e.g., %0.3f) python format string
    attribute : string, optional
        Bottom-level attribute of ``value`` to be formatted.

    Returns
    -------
    formatted_value : string
        The formatted version of ``value.<attribute>``. If
        ``value.<attribute>`` is None, "--" is returned instead.

    Examples
    --------
    >>> stringify(None, '%s')
    '--'

    >>> stringify(1.2, '%0.3f')
    '1.200'

    """

    if attribute is not None and value is not None:
        quantity = nested_getattr(value, attribute)
    else:
        quantity = value

    if quantity is None:
        return "--"
    else:
        return fmt % quantity


def classifier(value, bins, units=None):
    """
    An example classifier function for `storm_histogram`

    Parameters
    ----------
    value : float
        Valueue to classify.
    bins : array-like
        Finite right-edges of the classification bins.
    units : string, optional
        Units of measure to be appended to the returned categories.

    Returns
    -------
    category : string

    Examples
    --------
    >>> bins = [5, 10, 15, 20, 25]
    >>> classifier(3, bins)
    '<5'

    >>> classifier(12, bins, units='feet')
    '10 - 15 feet'

    >>> classifier(48, bins, units='mm')
    '>25 mm'

    """

    units = units or ""

    if numpy.isnan(value):
        return numpy.nan

    # below the lower edge
    elif value <= min(bins):
        output = f"<{min(bins)}"

    # above the upper edge
    elif value > max(bins):
        output = f">{max(bins)}"

    # everything else with the range of bins
    else:
        for left, right in zip(bins[:-1], bins[1:]):
            if left < value <= right:
                output = f"{left} - {right}"
                break

    # add the units (or don't)
    return "{} {}".format(output, units or "").strip()


def unique_categories(classifier, bins):
    """
    Computs all of the unique category returned by a classifier.

    Parameters
    ----------
    classifier : callable
        A function or class that we called categorizes continuous
        (floating point) values.
    bins : array-like
        Finite right-edges of the classification bins.

    Returns
    -------
    categories : list
        List, sequentially ordered with the possible input values, of
        all of the possible categories returned from ``classifier``.

    Examples
    --------
    >>> from functools import partial
    >>> from wqio import utils
    >>> bins = [5, 10, 15]
    >>> classifier = partial(utils.classifier, bins=bins, units='mm')
    >>> utils.unique_categories(classifier, bins)
    ['<5 mm', '5 - 10 mm', '10 - 15 mm', '>15 mm']

    """

    bins = numpy.asarray(bins)
    midpoints = 0.5 * (bins[:-1] + bins[1:])
    all_bins = [min(bins) * 0.5] + list(midpoints) + [max(bins) * 2]
    return [classifier(value) for value in all_bins]


def pop_many(some_dict, *args):
    """Pop several key-values out of a dictionary and return a copy

    Parameters
    ----------
    some_dict : dictionary
    *keys : hashables
        All of the keys you would like removed from *some_dict*

    Returns
    -------
    popped : dictionary

    Example
    -------
    >>> from wqio import utils
    >>> x = {'A': 1, 'B': 2, 'C': 3, 'D': 4}
    >>> utils.pop_many(x, 'A', 'D')
    {'B': 2, 'C': 3}

    """

    popped = copy(some_dict)
    for v in args:
        _ = popped.pop(v)
    return popped


def selector(default, *cond_results):
    """Thin wrapper around numpy.select with a more convenient API (maybe).

    Parameters
    ----------
    default : scalar
        The default value to be returned when none of the conditions are met.abs
    *cond_results : tuple
        Tuples of conditions (bool arrays) and result values (scalar)

    Returns
    -------
    selected : numpy.array

    Example
    -------
    >>> from wqio import utils
    >>> import numpy
    >>> x = numpy.arange(10)
    >>> utils.selector('Z', (x <= 2, 'A'), (x < 6, 'B'), (x <= 7, 'C'))
    array(['A', 'A', 'A', 'B', 'B', 'B', 'C', 'C', 'Z', 'Z'], dtype='<U1')
    """
    conditions, results = zip(*cond_results)
    return numpy.select(conditions, results, default)


def non_filter(*args, **kwargs):
    return True


def no_op(value):
    return value


def assign_multilevel_column(df, val_or_fxn, *collevels):
    """Dataframe-pipeable function to assign new multi-level columns

    Parameters
    ----------
    df : pandas.DataFrame
        Dataframe with a multilevel index along the column axis.
    val_or_fxn : scalar, array-like or callable
        If scalar array-like, theses are values that will be assigned directly.
        If this is a callable, the values returned by ``val_or_fxn(df)`` will be
        used.
    *collevels
        All values needed to specify the levels of the new column.

    Returns
    -------
    assigned : pandas.DataFrame
        Copy of the original dataframe with the new value assigned.

    """

    df = df.copy()
    if callable(val_or_fxn):
        df[collevels] = val_or_fxn(df)
    else:
        df[collevels] = val_or_fxn
    return df


def symbolize_bools(df, true_symbol, false_symbol, other_symbol=None, join_char=None):
    """Symbolize boolean values in a dataframe with strings

    Parameters
    ----------
    df : pandas.DataFrame
        A dataframe of boolean values
    true_symbol, false_symbol : str
        Strings used to symbolize True and False values, respectively
    other_symbol : str, optional
        String used to symbolized values other than True or False. If omitted,
        *false_symbol* is used.
    join_char : str, optional
        If provided, this character is used to join all of the columns in the
        dataframe into a single series. Otherwise, a dataframe is returned

    Returns
    -------
    symbolized : pandas.DataFrame or pandas.Series

    Examples
    --------
    >>> from numpy import nan
    >>> from pandas import DataFrame
    >>> from wqio.utils import symbolize_bools
    >>> df = DataFrame({
    ...     'A': [True, False, False],
    ...     'B': [False, True, True],
    ...     'C': [False, True, nan]
    ... })
    >>> symbolize_bools(df, true_symbol='◆', false_symbol='◇',
    ...                 other_symbol='✖', join_char='-')
    0    ◆-◇-◇
    1    ◇-◆-◆
    2    ◇-◆-✖
    dtype: object

    """

    if other_symbol is None:
        other_symbol = false_symbol

    mapper = {True: true_symbol, False: false_symbol}
    symbolized = df.map(lambda x: mapper.get(x, other_symbol))
    if join_char is None:
        return symbolized
    return symbolized.apply(lambda r: join_char.join(r), axis=1)


def log_df_shape(logger):  # pragma: no cover
    """Decorator to log the shape of a dataframe before and after a function."""

    def decorate(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            shape_init = args[0].shape
            new_df = func(*args, **kwargs)
            shape_final = new_df.shape
            logger.debug(f"{func.__name__}: dataframe shape = {shape_init} -> {shape_final}.")
            return new_df

        return wrapper

    return decorate


def log_or_warn(msg, warning=None, logger=None):
    if warning:
        warnings.warn(msg, warning)

    if logger:
        logger(msg)



import pandas
import pytest
from matplotlib import pyplot

from wqio import samples
from wqio.tests import helpers

BASELINE_IMAGES = "_baseline_images/samples_tests"
TOLERANCE = helpers.get_img_tolerance()


@pytest.fixture
def sample_data():
    datafile = helpers.test_data_path("test_wqsample_data.csv")
    rawdata = pandas.read_csv(datafile, index_col=[0, 1, 2, 3, 4, 5, 6, 11, 12])
    return rawdata


@pytest.fixture
def xlims():
    xlims = (
        pandas.Timestamp("2013-02-24 16:45").to_pydatetime(),
        pandas.Timestamp("2013-02-25 03:30").to_pydatetime(),
    )
    return xlims


@pytest.fixture
def grab_sample(sample_data):
    starttime = "2013-02-24 16:54"
    endtime = "2013-02-24 16:59"
    freq = None

    wqs = samples.GrabSample(sample_data, starttime, endtime=endtime, samplefreq=freq, storm=None)
    wqs.marker = "D"
    wqs.markersize = 8

    return wqs


@pytest.fixture
def composite_sample(sample_data):
    starttime = "2013-02-24 16:59"
    endtime = "2013-02-25 02:59"
    freq = pandas.offsets.Minute(20)

    wqs = samples.CompositeSample(
        sample_data, starttime, endtime=endtime, samplefreq=freq, storm=None
    )
    wqs.marker = "D"
    wqs.markersize = 8

    return wqs


@pytest.mark.parametrize(
    ("param", "unit", "comma", "expected"),
    [
        ("Copper", "ug/L", False, "Copper (ug/L)"),
        ("Copper", "ug/L", True, "Copper, ug/L"),
        ("Nitrate & Nitrite", "mg/L", False, "Nitrate & Nitrite (mg/L)"),
        ("Nitrate & Nitrite", "mg/L", True, "Nitrate & Nitrite, mg/L"),
    ],
)
def test_parameter_unit(param, unit, comma, expected):
    parameter = samples.Parameter(name=param, units=unit)
    assert parameter.paramunit(usecomma=comma) == expected


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_grabsample_noStorm_notFocus(grab_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    grab_sample.plot_ts(ax, isFocus=False)
    return fig


@pytest.mark.xfail
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_grabsample_noStorm_notFocus_asRug(grab_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    grab_sample.plot_ts(ax, isFocus=False, asrug=True)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_grabsample_noStorm_asFocus(grab_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    grab_sample.plot_ts(ax, isFocus=True)
    return fig


@pytest.mark.xfail
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_grabsample_noStorm_Focus_asRug(grab_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    grab_sample.plot_ts(ax, isFocus=True, asrug=True)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_compsample_noStorm_notFocus(composite_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    composite_sample.plot_ts(ax, isFocus=False)
    return fig


@pytest.mark.xfail
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_compsample_noStorm_notFocus_asRug(composite_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    composite_sample.plot_ts(ax, isFocus=False, asrug=True)
    return fig


@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_compsample_noStorm_Focus(composite_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    composite_sample.plot_ts(ax, isFocus=True)
    return fig


@pytest.mark.xfail
@pytest.mark.mpl_image_compare(baseline_dir=BASELINE_IMAGES, tolerance=TOLERANCE)
def test_plot_compsample_noStorm_Focus_asRug(composite_sample, xlims):
    fig, ax = pyplot.subplots()
    ax.set_xlim(xlims)
    composite_sample.plot_ts(ax, isFocus=True, asrug=True)
    return fig



from wqio import bootstrap  # noqa
from wqio import utils  # noqa
from wqio.datacollections import DataCollection  # noqa
from wqio.datasets import download  # noqa
from wqio.features import Dataset, Location  # noqa
from wqio.hydro import DrainageArea, HydroRecord, Storm, parse_storm_events  # noqa
from wqio.ros import ROS, cohn_numbers  # noqa
from wqio.samples import CompositeSample, GrabSample, Parameter  # noqa
from wqio.tests import test, teststrict  # noqa



import warnings
from collections import namedtuple
from functools import partial

import numpy
import pandas
import statsmodels.api as sm
from scipy import stats
from statsmodels.tools.decorators import cache_readonly

try:
    from tqdm import tqdm
except ImportError:  # pragma: no cover
    tqdm = None

from wqio import bootstrap, utils, validate
from wqio.features import Dataset, Location
from wqio.ros import ROS

_Stat = namedtuple("_stat", ["stat", "pvalue"])


def _dist_compare(x, y, stat_comp_func):
    if (len(x) == len(y)) and numpy.equal(x, y).all():
        return _Stat(numpy.nan, numpy.nan)
    return stat_comp_func(x, y, alternative="two-sided")


class DataCollection:
    """Generalized water quality comparison object.

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Dataframe all of the data to analyze.
    rescol, qualcol, stationcol, paramcol : string
        Column labels for the results, qualifiers, stations (monitoring
        locations), and parameters (pollutants), respectively.

        .. note::

           Non-detect results should be reported as the detection
           limit of that observation.

    ndval : string or list of strings, options
        The values found in ``qualcol`` that indicates that a
        result is a non-detect.
    othergroups : list of strings, optional
        The columns (besides ``stationcol`` and ``paramcol``) that
        should be considered when grouping into subsets of data.
    pairgroups : list of strings, optional
        Other columns besides ``stationcol`` and ``paramcol`` that
        can be used define a unique index on ``dataframe`` such that it
        can be "unstack" (i.e., pivoted, cross-tabbed) to place the
        ``stationcol`` values into columns. Values of ``pairgroups``
        may overlap with ``othergroups``.
    useros : bool (default = True)
        Toggles the use of regression-on-order statistics to estimate
        non-detect values when computing statistics.
    filterfxn : callable, optional
        Function that will be passed to the ``filter`` method of a
        ``pandas.Groupby`` object to remove groups that should not be
        analyzed (for whatever reason). If not provided, all groups
        returned by ``dataframe.groupby(by=groupcols)`` will be used.
    bsiter : int
        Number of iterations the bootstrapper should use when estimating
        confidence intervals around a statistic.
    showpbar : bool (True)
        When True and the `tqdm` module is available, this will toggle the
        appears of progress bars in long-running group by-apply operations.

    """

    # column that stores the censorsip status of an observation
    cencol = "__censorship"

    def __init__(
        self,
        dataframe,
        rescol="res",
        qualcol="qual",
        stationcol="station",
        paramcol="parameter",
        ndval="ND",
        othergroups=None,
        pairgroups=None,
        useros=True,
        filterfxn=None,
        bsiter=10000,
        showpbar=True,
    ):
        # cache for all of the properties
        self._cache = {}

        # basic input
        self.raw_data = dataframe
        self._raw_rescol = rescol
        self.qualcol = qualcol
        self.stationcol = stationcol
        self.paramcol = paramcol
        self.ndval = validate.at_least_empty_list(ndval)
        self.othergroups = validate.at_least_empty_list(othergroups)
        self.pairgroups = validate.at_least_empty_list(pairgroups)
        self.useros = useros
        self.filterfxn = filterfxn or utils.non_filter
        self.bsiter = bsiter
        self.showpbar = showpbar

        # column that stores ROS'd values
        self.roscol = "ros_" + rescol

        # column stators "final" values
        if self.useros:
            self.rescol = self.roscol
        else:
            self.rescol = rescol

        # columns to group by when ROS'd, doing general stats
        self.groupcols = [self.stationcol, self.paramcol] + self.othergroups
        self.groupcols_comparison = [self.paramcol] + self.othergroups

        # columns to group and pivot by when doing paired stats
        self.pairgroups = self.pairgroups + [self.stationcol, self.paramcol]

        # final column list of the tidy dataframe
        self.tidy_columns = self.groupcols + [self._raw_rescol, self.cencol]

        # the "raw" data with the censorship column added
        self.data = dataframe.assign(
            **{self.cencol: dataframe[self.qualcol].isin(self.ndval)}
        ).reset_index()

    @cache_readonly
    def tidy(self):
        if self.useros:

            def fxn(g):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    rosdf = (
                        ROS(
                            df=g,
                            result=self._raw_rescol,
                            censorship=self.cencol,
                            as_array=False,
                        )
                        .rename(columns={"final": self.roscol})
                        .loc[:, [self._raw_rescol, self.roscol, self.cencol]]
                    )
                return rosdf

        else:

            def fxn(g):
                g[self.roscol] = numpy.nan
                return g

        if tqdm and self.showpbar:

            def make_tidy(df):
                tqdm.pandas(desc="Tidying the DataCollection")
                return df.groupby(self.groupcols).progress_apply(fxn, include_groups=False)

        else:

            def make_tidy(df):
                return df.groupby(self.groupcols).apply(fxn, include_groups=False)

        keep_cols = self.tidy_columns + [self.roscol]
        with warnings.catch_warnings():
            warnings.simplefilter("once")
            _tidy = (
                self.data.reset_index()[self.tidy_columns]
                .groupby(by=self.groupcols)
                .filter(self.filterfxn)
                .pipe(make_tidy)
                .reset_index()
                .sort_values(by=self.groupcols)
            )

        return _tidy[keep_cols]

    @cache_readonly
    def paired(self):
        _pairs = (
            self.data.reset_index()
            .groupby(by=self.groupcols)
            .filter(self.filterfxn)
            .set_index(self.pairgroups)
            .unstack(level=self.stationcol)
            .rename_axis(["value", self.stationcol], axis="columns")
        )[[self._raw_rescol, self.cencol]]
        return _pairs

    def generic_stat(
        self,
        statfxn,
        use_bootstrap=True,
        statname=None,
        has_pvalue=False,
        filterfxn=None,
        **statopts,
    ):
        """Generic function to estimate a statistic and its CIs.

        Parameters
        ----------
        statfxn : callable
            A function that takes a 1-D sequnce and returns a scalar
            results. Its call signature should be in the form:
            ``statfxn(seq, **kwargs)``.
        use_bootstrap : bool, optional
            Toggles using a BCA bootstrapping method to estimate the
            95% confidence interval around the statistic.
        statname : string, optional
            Name of the statistic. Included as a column name in the
            final dataframe.
        has_pvalue : bool, optional
            Set to ``True`` if ``statfxn`` returns a tuple of the
            statistic and it's p-value.
        **statopts : optional kwargs
            Additional keyword arguments that will be passed to
            ``statfxn``.

        Returns
        -------
        stat_df : pandas.DataFrame
            A dataframe all the results of the ``statfxn`` when applied
            to ``self.tidy.groupby(self.groupcols)``.

        Examples
        --------
        This actually demonstrates how ``DataCollection.mean`` is
        implemented.

        >>> import numpy
        >>> import wqio
        >>> from wqio.tests import helpers
        >>> df = helpers.make_dc_data_complex()
        >>> dc = DataCollection(df, rescol='res', qualcol='qual',
        ...                     stationcol='loc', paramcol='param',
        ...                     ndval='<')
        >>> means = dc.generic_stat(numpy.mean, statname='Arith. Mean')

        You can also use ``lambda`` objects

        >>> pctl35 = dc.generic_stat(lambda x: numpy.percentile(x, 35),
        ...                          statname='pctl35', use_bootstrap=False)

        """

        if statname is None:
            statname = "stat"

        if filterfxn is None:
            filterfxn = utils.non_filter

        def fxn(x):
            data = x[self.rescol].values
            if use_bootstrap:
                stat = statfxn(data)
                lci, uci = bootstrap.BCA(data, statfxn=statfxn)
                values = [lci, stat, uci]
                statnames = ["lower", statname, "upper"]
            else:
                values = validate.at_least_empty_list(statfxn(data, **statopts))
                if hasattr(values, "_fields"):  # nametuple
                    statnames = values._fields
                else:  # tuple
                    statnames = [statname]
                    if has_pvalue:
                        statnames.append("pvalue")

            return pandas.Series(values, index=statnames)

        groups = self.tidy.groupby(by=self.groupcols).filter(filterfxn).groupby(by=self.groupcols)

        if tqdm and self.showpbar:
            tqdm.pandas(desc="Computing stats")
            vals = groups.progress_apply(fxn, include_groups=False)
        else:
            vals = groups.apply(fxn, include_groups=False)

        results = (
            vals.unstack(level=self.stationcol)
            .pipe(utils.swap_column_levels, 0, 1)
            .rename_axis(["station", "result"], axis="columns")
        )
        return results

    @cache_readonly
    def count(self):
        return (
            self.generic_stat(lambda x: x.shape[0], use_bootstrap=False, statname="Count")
            .fillna(0)
            .astype(int)
        )

    @cache_readonly
    def inventory(self):
        counts = (
            self.tidy.groupby(by=self.groupcols + [self.cencol])
            .size()
            .unstack(level=self.cencol)
            .fillna(0)
            .astype(int)
            .rename_axis(None, axis="columns")
            .rename(columns={False: "Detect", True: "Non-Detect"})
            .assign(Count=lambda df: df.sum(axis="columns"))
        )
        if "Non-Detect" not in counts.columns:
            counts["Non-Detect"] = 0

        return counts[["Count", "Non-Detect"]]

    @cache_readonly
    def median(self):
        return self.generic_stat(numpy.median, statname="median")

    @cache_readonly
    def mean(self):
        return self.generic_stat(numpy.mean, statname="mean")

    @cache_readonly
    def std_dev(self):
        return self.generic_stat(numpy.std, statname="std. dev.", use_bootstrap=False)

    def percentile(self, percentile):
        """Return the percentiles (0 - 100) for the data."""
        return self.generic_stat(
            lambda x: numpy.percentile(x, percentile),
            statname=f"pctl {percentile}",
            use_bootstrap=False,
        )

    @cache_readonly
    def logmean(self):
        return self.generic_stat(
            lambda x, axis=0: numpy.mean(numpy.log(x), axis=axis), statname="Log-mean"
        )

    @cache_readonly
    def logstd_dev(self):
        return self.generic_stat(
            lambda x, axis=0: numpy.std(numpy.log(x), axis=axis),
            use_bootstrap=False,
            statname="Log-std. dev.",
        )

    @cache_readonly
    def geomean(self):
        geomean = numpy.exp(self.logmean)
        geomean.columns.names = ["station", "Geo-mean"]
        return geomean

    @cache_readonly
    def geostd_dev(self):
        geostd = numpy.exp(self.logstd_dev)
        geostd.columns.names = ["station", "Geo-std. dev."]
        return geostd

    @cache_readonly
    def shapiro(self):
        return self.generic_stat(
            stats.shapiro,
            use_bootstrap=False,
            has_pvalue=True,
            statname="shapiro",
            filterfxn=lambda x: x.shape[0] > 3,
        )

    @cache_readonly
    def shapiro_log(self):
        return self.generic_stat(
            lambda x: stats.shapiro(numpy.log(x)),
            use_bootstrap=False,
            has_pvalue=True,
            filterfxn=lambda x: x.shape[0] > 3,
            statname="log-shapiro",
        )

    @cache_readonly
    def lilliefors(self):
        return self.generic_stat(
            sm.stats.lilliefors,
            use_bootstrap=False,
            has_pvalue=True,
            statname="lilliefors",
        )

    @cache_readonly
    def lilliefors_log(self):
        return self.generic_stat(
            lambda x: sm.stats.lilliefors(numpy.log(x)),
            use_bootstrap=False,
            has_pvalue=True,
            statname="log-lilliefors",
        )

    @cache_readonly
    def anderson_darling(self):
        raise NotImplementedError
        return self.generic_stat(
            utils.anderson_darling,
            use_bootstrap=False,
            has_pvalue=True,
            statname="anderson-darling",
        )

    @cache_readonly
    def anderson_darling_log(self):
        raise NotImplementedError
        return self.generic_stat(
            lambda x: utils.anderson_darling(numpy.log(x)),
            use_bootstrap=False,
            has_pvalue=True,
            statname="log-anderson-darling",
        )

    def comparison_stat(self, statfxn, statname=None, paired=False, **statopts):
        """Generic function to apply comparative hypothesis tests to
        the groups of the ``DataCollection``.

        Parameters
        ----------
        statfxn : callable
            A function that takes a 1-D sequnce and returns a scalar
            results. Its call signature should be in the form:
            ``statfxn(seq, **kwargs)``.
        statname : string, optional
            Name of the statistic. Included as a column name in the
            final dataframe.
        apired : bool, optional
            Set to ``True`` if ``statfxn`` requires paired data.
        **statopts : optional kwargs
            Additional keyword arguments that will be passed to
            ``statfxn``.

        Returns
        -------
        stat_df : pandas.DataFrame
            A dataframe all the results of the ``statfxn`` when applied
            to ``self.tidy.groupby(self.groupcols)`` or
            ``self.paired.groupby(self.groupcols)`` when necessary.

        Examples
        --------
        This actually demonstrates how ``DataCollection.mann_whitney``
        is implemented.

        >>> from scipy import stats
        >>> import wqio
        >>> from wqio.tests import helpers
        >>> df = helpers.make_dc_data_complex()
        >>> dc = DataCollection(df, rescol='res', qualcol='qual',
        ...                     stationcol='loc', paramcol='param',
        ...                     ndval='<')
        >>> mwht = dc.comparison_stat(stats.mannwhitneyu,
        ...                           statname='mann_whitney',
        ...                           alternative='two-sided')

        """

        if paired:
            data = self.paired
            generator = utils.numutils._paired_stat_generator
            rescol = self._raw_rescol
        else:
            data = self.tidy
            generator = utils.numutils._comp_stat_generator
            rescol = self.rescol

        station_columns = [self.stationcol + "_1", self.stationcol + "_2"]
        meta_columns = self.groupcols_comparison
        index_cols = meta_columns + station_columns

        results = generator(
            data, meta_columns, self.stationcol, rescol, statfxn, statname=statname, **statopts
        )
        return pandas.DataFrame.from_records(results).set_index(index_cols)

    @cache_readonly
    def mann_whitney(self):
        return self.comparison_stat(
            partial(_dist_compare, stat_comp_func=stats.mannwhitneyu),
            statname="mann_whitney",
        )

    @cache_readonly
    def ranksums(self):
        return self.comparison_stat(stats.ranksums, statname="rank_sums")

    @cache_readonly
    def t_test(self):
        return self.comparison_stat(stats.ttest_ind, statname="t_test", equal_var=False)

    @cache_readonly
    def levene(self):
        return self.comparison_stat(stats.levene, statname="levene", center="median")

    @cache_readonly
    def wilcoxon(self):
        return self.comparison_stat(
            partial(_dist_compare, stat_comp_func=stats.wilcoxon),
            statname="wilcoxon",
            paired=True,
        )

    @cache_readonly
    def kendall(self):
        return self.comparison_stat(stats.kendalltau, statname="kendalltau", paired=True)

    @cache_readonly
    def spearman(self):
        return self.comparison_stat(stats.spearmanr, statname="spearmanrho", paired=True)

    @cache_readonly
    def theilslopes(self, logs=False):
        raise NotImplementedError

    @cache_readonly
    def locations(self):
        _locations = []
        groups = (
            self.data.groupby(by=self.groupcols).filter(self.filterfxn).groupby(by=self.groupcols)
        )
        cols = [self._raw_rescol, self.qualcol]
        for names, data in groups:
            loc_dict = dict(zip(self.groupcols, names))
            loc = (
                data.set_index(self.pairgroups)[cols]
                .reset_index(level=self.stationcol, drop=True)
                .pipe(
                    Location,
                    station_type=loc_dict[self.stationcol].lower(),
                    rescol=self._raw_rescol,
                    qualcol=self.qualcol,
                    ndval=self.ndval,
                    bsiter=self.bsiter,
                    useros=self.useros,
                )
            )

            loc.definition = loc_dict
            _locations.append(loc)

        return _locations

    def datasets(self, loc1, loc2):
        """Generate ``Dataset`` objects from the raw data of the
        ``DataColletion``.

        Data are first grouped by ``self.groupcols`` and
        ``self.stationcol``. Data frame each group are then queried
        for into separate ``Lcoations`` from ``loc1`` and ``loc2``.
        The resulting ``Locations`` are used to create a ``Dataset``.

        Parameters
        ----------
        loc1, loc2 : string
            Values found in the ``self.stationcol`` property that will
            be used to distinguish the two ``Location`` objects for the
            ``Datasets``.

        Yields
        ------
        ``Dataset`` objects

        """

        groupcols = list(filter(lambda g: g != self.stationcol, self.groupcols))

        for names, data in self.data.groupby(by=groupcols):
            ds_dict = dict(zip(groupcols, names))

            ds_dict[self.stationcol] = loc1
            infl = self.selectLocations(squeeze=True, **ds_dict)

            ds_dict[self.stationcol] = loc2
            effl = self.selectLocations(squeeze=True, **ds_dict)

            ds_dict.pop(self.stationcol)
            dsname = "_".join(names).replace(", ", "")

            if effl:
                ds = Dataset(infl, effl, useros=self.useros, name=dsname)
                ds.definition = ds_dict
                yield ds

    @staticmethod
    def _filter_collection(collection, squeeze, **kwargs):
        items = list(collection)
        for key, value in kwargs.items():
            if numpy.isscalar(value):
                items = [r for r in filter(lambda x: x.definition[key] == value, items)]
            else:
                items = [r for r in filter(lambda x: x.definition[key] in value, items)]

        if squeeze:
            if len(items) == 1:
                items = items[0]
            elif len(items) == 0:
                items = None

        return items

    def selectLocations(self, squeeze=False, **conditions):
        """Select ``Location`` objects meeting specified criteria
        from the ``DataColletion``.

        Parameters
        ----------
        squeeze : bool, optional
            When True and only one object is found, it returns the bare
            object. Otherwise, a list is returned.
        **conditions : optional parameters
            The conditions to be applied to the definitions of the
            ``Locations`` to filter them out. If a scalar is provided
            as the value, normal comparison (==) is used. If a sequence
            is provided, the ``in`` operator is used.

        Returns
        -------
        locations : list of ``wqio.Location`` objects

        Example
        -------
        >>> from wqio.tests.helpers import make_dc_data_complex
        >>> import wqio
        >>> df = make_dc_data_complex()
        >>> dc = wqio.DataCollection(df, rescol='res', qualcol='qual',
        ...                          stationcol='loc', paramcol='param',
        ...                          ndval='<', othergroups=None,
        ...                          pairgroups=['state', 'bmp'],
        ...                          useros=True, bsiter=10000)
        >>> locs = dc.selectLocations(param=['A', 'B'], loc=['Inflow', 'Reference'])
        >>> len(locs)
        4
        >>> locs[0].definition
        {'loc': 'Inflow', 'param': 'A'}

        """

        locations = self._filter_collection(self.locations.copy(), squeeze=squeeze, **conditions)
        return locations

    def selectDatasets(self, loc1, loc2, squeeze=False, **conditions):
        """Select ``Dataset`` objects meeting specified criteria
        from the ``DataColletion``.

        Parameters
        ----------
        loc1, loc2 : string
            Values found in the ``self.stationcol`` property that will
            be used to distinguish the two ``Location`` objects for the
            ``Datasets``.
        squeeze : bool, optional
            When True and only one object is found, it returns the bare
            object. Otherwise, a list is returned.
        **conditions : optional parameters
            The conditions to be applied to the definitions of the
            ``Locations`` to filter them out. If a scalar is provided
            as the value, normal comparison (==) is used. If a sequence
            is provided, the ``in`` operator is used.

        Returns
        -------
        locations : list of ``wqio.Location`` objects

        Example
        -------
        >>> from wqio.tests.helpers import make_dc_data_complex
        >>> import wqio
        >>> df = make_dc_data_complex()
        >>> dc = wqio.DataCollection(df, rescol='res', qualcol='qual',
        ...                          stationcol='loc', paramcol='param',
        ...                          ndval='<', othergroups=None,
        ...                          pairgroups=['state', 'bmp'],
        ...                          useros=True, bsiter=10000)
        >>> dsets = dc.selectDatasets('Inflow', 'Outflow', squeeze=False,
        ... param=['A', 'B'])
        >>> len(dsets)
        2
        >>> dsets[0].definition
        {'param': 'A'}
        """

        datasets = self._filter_collection(self.datasets(loc1, loc2), squeeze=squeeze, **conditions)
        return datasets

    def n_unique(self, column):
        return (
            self.data.loc[:, self.groupcols + [column]]
            .drop_duplicates()
            .groupby(self.groupcols)
            .size()
            .unstack(level=self.stationcol)
            .pipe(utils.add_column_level, column, "result")
            .swaplevel(axis="columns")
            .fillna(0)
            .astype(int)
        )

    def stat_summary(self, percentiles=None, groupcols=None, useros=True):
        """A generic, high-level summary of the data collection.

        Parameters
        ----------
        groupcols : list of strings, optional
            The columns by which ``self.tidy`` will be grouped when
            computing the statistics.
        useros : bool, optional
            Toggles of the use of the ROS'd (``True``) or raw
            (``False``) data.

        Returns
        -------
        stat_df : pandas.DataFrame

        """

        col = self.roscol if useros else self.rescol
        groupcols = validate.at_least_empty_list(groupcols) or self.groupcols
        ptiles = percentiles or [0.1, 0.25, 0.5, 0.75, 0.9]
        summary = (
            self.tidy.groupby(by=groupcols)
            .apply(lambda g: g[col].describe(percentiles=ptiles).T)
            .drop("count", axis="columns")
        )
        return self.inventory.join(summary).unstack(level=self.stationcol)



import numpy
import pandas
import seaborn
import statsmodels.api as sm
from matplotlib import pyplot
from probscale.algo import _estimate_from_fit
from scipy import stats
from statsmodels.tools.decorators import cache_readonly

from wqio import bootstrap, utils, validate, viz
from wqio.ros import ROS

# meta data mappings based on station
station_names = {
    "inflow": "Influent",
    "outflow": "Effluent",
    "reference": "Reference Flow",
}

markers = {"Influent": ["o", "v"], "Effluent": ["s", "<"], "Reference Flow": ["D", "d"]}

palette = seaborn.color_palette(palette="deep", n_colors=3, desat=0.88)
colors = {"Influent": palette[0], "Effluent": palette[1], "Reference Flow": palette[2]}


class Location:
    """Object providing convenient access to statistical and
    graphical methods for summarizing a single set of water quality
    observations for a single pollutant.

    Parameters
    -----------
    dataframe : pandas.DataFrame
        A dataframe that contains at least two columns: one for the
        analytical values and another for the data qualfiers. Can
        contain any type of row index, but the column index must be
        simple (i.e., not a pandas.MultiIndex).
    rescol : string, optional (default = 'res')
        Name of the column in `dataframe` that contains the
        analytical values.
    qualcol : string, optional (default = 'qual')
        Name of the column in `dataframe` containing qualifiers.
    ndval : string, optional (default = 'ND')
        The *only* value in the `qualcol` of `dataframe` that
        indicates that the corresponding value in `rescol` is
        non-detect.
    station_type : string, optional
        Type of location being analyzed. Valid values are:
        'inflow' (default) or 'outflow'.
    useros : bool, optional (default = True)
        Toggles the use of Regression On Order Statistics to
        estimate non-detect values when computing statistics.
    cencol : string, optional (default = 'cen')
        Name of the column indicaticating if a results is censored.
        These values will be computed from ``qualcol`` and ``ndval``.
    bsiter : int, optional (default = 1e4)
        Number of interations to use when using a bootstrap
        algorithm to refine a statistic.
    include : bool, optional (default = True)
        Toggles the inclusion of the location when programmatically
        creating many `Location` objects.

    Settable Properties
    -------------------
    .name : string
        A human-readable name for the data.
    .definition : dict
        A dictionary of key-value pairs that define what makes this
        data distinct within a larger collection of `Location` objects.
    .include : bool
        Same as input.

    Statistical Attributes
    ----------------------
    .N : int
        Total number of results.
    .ND : int
        Number of non-detect results.
    .NUnique : int
        Number of unique result values in the data.
    .fractionND : float
        Fraction of data that is non-detect.
    .min : float
        Minimum value of the data.
    .max : float
        Maximum value of the data.
    .min_detect : float
        Minimum detected value of the data.
    .min_DL : float
        Minimum detection limit reported for the data.
    .mean*+^ : float
        Bootstrapped arithmetic mean of the data.
    .std*+^ : float
        Bootstrapped standard deviation of the data.
    .cov : float
        Covariance (absolute value of std/mean).
    .skew : float
        Skewness coefficient.
    .median* : float
        Median of the dataset.
    .pctl[10/25/75/90] : float
        Percentiles of the dataset.
    .pnorm : float
        Results of the Shapiro-Wilks test for normality.
    .plognorm : float
        Results of the Shapiro-Wilks test for normality on
        log-transormed data (so really a test for lognormalily).
    .lilliefors+ : list of floats
        Lilliefors statistic and p-value.
    .shapiro+ : list of floats
        Shapiro-Wilks statistic and p-value.
    .anderson+ : tuple
        Anderson-Darling statistic, critical values, significance
        levels.
    .analysis_space : string
        Based on the results of self.pnorm and self.plognorm, this is
        either "normal" or "lognormal".

    Statistial Notes
    ----------------
    * Indicates that there's an accompyaning tuple of confidence
      interval. For example, self.mean and self.mean_conf_interval
    + Indicatates that there's a equivalent stat for log-transormed
      data. For example, self.mean and self.logmean or self.lilliefors
      and self.lilliefors_log (subject to the absense of negitive
      results).
    ^ Indicatates that there's a equivalent stat in geometric space.
      For example, self.mean and self.geomean.

    Plotting Methods
    ----------------
    .verticalScatter
    .boxplot
    .probplot
    .statplot

    """

    def __init__(
        self,
        dataframe,
        rescol="res",
        qualcol="qual",
        ndval="ND",
        station_type="inflow",
        useros=True,
        cencol="cen",
        bsiter=10000,
        include=True,
    ):
        # plotting symbology based on location type
        self.station_type = station_type
        self.station_name = station_names[station_type]
        self.plot_marker = markers[self.station_name][0]
        self.scatter_marker = markers[self.station_name][1]
        self.color = colors[self.station_name]

        # basic stuff
        self._name = self.station_name
        self._include = include
        self._definition = {}

        # parameters of the stats analysis
        self._cache = {}

        # properties of the dataframe and analysis
        self.bsiter = bsiter
        self.useros = useros
        self.rescol = rescol
        self.qualcol = qualcol
        self.cencol = cencol
        if numpy.isscalar(ndval):
            self.ndvals = [ndval]
        else:
            self.ndvals = ndval

        # original data and quantity
        self.raw_data = dataframe.assign(**{self.cencol: dataframe[qualcol].isin(self.ndvals)})
        self._dataframe = None
        self._data = None

    @property
    def dataframe(self):
        if self.raw_data.shape[0] > 0 and self._dataframe is None:
            df = self.raw_data.assign(
                **{self.cencol: lambda df: df[self.qualcol].isin(self.ndvals)}
            )
            if self.useros:
                ros = ROS(df=df, result=self.rescol, censorship=self.cencol, as_array=False)
                self._dataframe = ros[["final", self.cencol]]
            else:
                self._dataframe = df[[self.rescol, self.cencol]]
        return self._dataframe

    @property
    @numpy.deprecate
    def full_data(self):
        return self.dataframe

    @property
    def data(self):
        if self.hasData:
            if self.useros:
                output = self.dataframe["final"].values
            else:
                output = self.dataframe[self.rescol].values

            return output

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, value):
        self._definition = value

    @property
    def include(self):
        return self._include

    @include.setter
    def include(self, value):
        self._include = value

    @property
    def exclude(self):
        return not self.include

    @cache_readonly
    def N(self):
        return self.data.shape[0]

    @cache_readonly
    def hasData(self):
        return self.dataframe.shape[0] > 0

    @cache_readonly
    def all_positive(self):
        if self.hasData:
            return self.min > 0

    @cache_readonly
    def ND(self):
        return self.dataframe[self.cencol].sum()

    @cache_readonly
    def NUnique(self):
        return pandas.unique(self.raw_data[self.rescol]).shape[0]

    @cache_readonly
    def fractionND(self):
        return self.ND / self.N

    @cache_readonly
    def shapiro(self):
        if self.hasData:
            return stats.shapiro(self.data)

    @cache_readonly
    def shapiro_log(self):
        if self.hasData:
            return stats.shapiro(numpy.log(self.data))

    @cache_readonly
    def lilliefors(self):
        if self.hasData:
            return sm.stats.lilliefors(self.data)

    @cache_readonly
    def lilliefors_log(self):
        if self.hasData:
            return sm.stats.lilliefors(numpy.log(self.data))

    @cache_readonly
    def anderson(self):
        if self.hasData:
            return utils.anderson_darling(self.data)

    @cache_readonly
    def anderson_log(self):
        if self.hasData:
            return utils.anderson_darling(numpy.log(self.data))

    @cache_readonly
    def analysis_space(self):
        if self.shapiro_log[1] >= self.shapiro[1] and self.shapiro_log[1] > 0.1:
            return "lognormal"
        else:
            return "normal"

    @cache_readonly
    def cov(self):
        if self.hasData:
            return self.data.std() / self.data.mean()

    @cache_readonly
    def min(self):
        if self.hasData:
            return self.data.min()

    @cache_readonly
    def min_detect(self):
        if self.hasData:
            return self.raw_data[self.rescol][~self.raw_data[self.qualcol].isin(self.ndvals)].min()

    @cache_readonly
    def min_DL(self):
        if self.hasData:
            return self.raw_data[self.rescol][self.raw_data[self.qualcol].isin(self.ndvals)].min()

    @cache_readonly
    def max(self):
        if self.hasData:
            return self.data.max()

    @cache_readonly
    def skew(self):
        if self.hasData:
            return stats.skew(self.data)

    @cache_readonly
    def pctl10(self):
        if self.hasData:
            return numpy.percentile(self.data, 10)

    @cache_readonly
    def pctl25(self):
        if self.hasData:
            return numpy.percentile(self.data, 25)

    @cache_readonly
    def pctl75(self):
        if self.hasData:
            return numpy.percentile(self.data, 75)

    @cache_readonly
    def pctl90(self):
        if self.hasData:
            return numpy.percentile(self.data, 90)

    # stats that we need
    @cache_readonly
    def median(self):
        if self.hasData:
            return numpy.median(self.data)

    @cache_readonly
    def median_conf_interval(self):
        if self.hasData:
            return bootstrap.BCA(self.data, numpy.median, niter=self.bsiter)

    @cache_readonly
    def mean(self):
        if self.hasData:
            return numpy.mean(self.data)

    @cache_readonly
    def mean_conf_interval(self):
        if self.hasData:
            return bootstrap.BCA(self.data, numpy.mean, niter=self.bsiter)

    @cache_readonly
    def std(self):
        if self.hasData:
            return numpy.std(self.data)

    @cache_readonly
    def logmean(self):
        if self.all_positive and self.hasData:
            return numpy.mean(numpy.log(self.data))

    @cache_readonly
    def logmean_conf_interval(self):
        if self.all_positive and self.hasData:

            def fxn(x, **kwds):
                return numpy.mean(numpy.log(x), **kwds)

            return bootstrap.BCA(self.data, fxn, niter=self.bsiter)

    @cache_readonly
    def logstd(self):
        if self.all_positive and self.hasData:
            return numpy.std(numpy.log(self.data))

    @cache_readonly
    def geomean(self):
        if self.all_positive and self.hasData:
            return numpy.exp(self.logmean)

    @cache_readonly
    def geomean_conf_interval(self):
        if self.all_positive and self.hasData:
            return numpy.exp(self.logmean_conf_interval)

    @cache_readonly
    def geostd(self):
        if self.all_positive and self.hasData:
            return numpy.exp(self.logstd)

    def boxplot_stats(self, log=True, bacteria=False):
        bxpstats = {
            "label": self.name,
            "mean": self.geomean if bacteria else self.mean,
            "med": self.median,
            "q1": self.pctl25,
            "q3": self.pctl75,
            "cilo": self.median_conf_interval[0],
            "cihi": self.median_conf_interval[1],
        }

        if log:
            wnf = viz.whiskers_and_fliers(
                numpy.log(self.data),
                numpy.log(self.pctl25),
                numpy.log(self.pctl75),
                transformout=numpy.exp,
            )
        else:
            wnf = viz.whiskers_and_fliers(self.data, self.pctl25, self.pctl75, transformout=None)

        bxpstats.update(wnf)
        return [bxpstats]

    # plotting methods
    def boxplot(
        self,
        ax=None,
        pos=1,
        yscale="log",
        shownotches=True,
        showmean=True,
        width=0.8,
        bacteria=False,
        ylabel=None,
        xlabel=None,
        patch_artist=False,
        xlims=None,
    ):
        """Draws a boxplot and whisker on a matplotlib figure

        Parameters
        ----------
        ax : optional matplotlib axes object or None (default)
            Axes on which the boxplot with be drawn. If None, one will
            be created.
        pos : optional int (default=1)
            Location along x-axis where boxplot will be placed.
        yscale : optional string ['linear' or 'log' (default)]
            Scale formatting of the y-axis
        shownotches : optional bool (default=True)
            Toggles drawing of bootstrapped confidence interval around
            the median.
        showmean : optional bool (default=True)
            Toggles plotting the mean value on the boxplot as a point.
            See also the `bacteria` kwarg
        width : optional float (default=0.8)
            Width of boxplot on the axes (data units)
        bacteria : optional bool (default False)
            If True, uses the geometric mean when `showmean` is True.
            Otherwise, the arithmetic mean is used.
        ylabel : string or None (default):
            Label for y-axis
        xlabel : string or None (default):
            Label for x-axis. If None, uses self.name
        patch_artist : optional bool (default = False)
            Toggles the use of patch artist instead of a line artists
            for the boxes
        xlims : dict, optional
            Dictionary of limits for the x-axis. Keys must be either
            "left", "right", or both

        Returns
        -------
        fig : matplotlib Figure

        """

        fig, ax = validate.axes(ax)
        y_log = yscale == "log"
        bxpstats = self.boxplot_stats(log=y_log, bacteria=bacteria)
        if xlabel is not None:
            bxpstats[0]["label"] = xlabel

        viz.boxplot(
            bxpstats,
            ax=ax,
            position=pos,
            width=width,
            color=self.color,
            marker=self.plot_marker,
            patch_artist=patch_artist,
            showmean=showmean,
            shownotches=shownotches,
        )

        ax.set_yscale(yscale)
        if yscale == "log":
            ax.yaxis.set_major_formatter(viz.log_formatter(use_1x=True, threshold=6))

        if ylabel:
            ax.set_ylabel(ylabel)

        ax.set_xticks([pos])
        if self.name is not None:
            ax.set_xticklabels([self.name])
        if xlabel is not None:
            ax.set_xticklabels([xlabel])

        if xlims is not None:
            ax.set_xlim(**xlims)

        return fig

    def probplot(
        self,
        ax=None,
        yscale="log",
        axtype="prob",
        ylabel=None,
        clearYLabels=False,
        rotateticklabels=True,
        bestfit=False,
        **plotopts,
    ):
        """Draws a probability plot on a matplotlib figure

        Parameters
        ----------
        ax : matplotlib axes object, optional or None (default).
            The Axes on which to plot. If None is provided, one will be
            created.
        axtype : string, optional (default = 'pp')
            Type of plot to be created. Options are:
                - 'prob': probabilty plot
                - 'pp': percentile plot
                - 'qq': quantile plot
        xlabel, ylabel : string, optional or None (default)
            Axis label for the plot.
        yscale : string , optional(default = 'log')
            Scale for the y-axis. Use 'log' for logarithmic (default) or
            'linear'.
        clearYLabels : bool, optional (default is False)
            If True, removed y-*tick* labels from teh Axes.
        rotateticklabels : bool, optional (default is True)
            If True, the tick labels of the probability axes will be
            rotated.
        bestfit : bool, optional (default is False)
            Specifies whether a best-fit line should be added to the
            plot.
        plotopts : keyword arguments
            Additional options passed directly to `pyplot.plot` when
            drawing the data points.

        Returns
        -------
        fig : matplotlib.Figure instance

        """

        fig, ax = validate.axes(ax)

        scatter_kws = plotopts.copy()
        scatter_kws["color"] = plotopts.get("color", self.color)
        scatter_kws["label"] = plotopts.get("label", self.name)
        scatter_kws["marker"] = plotopts.get("marker", self.plot_marker)
        scatter_kws["linestyle"] = plotopts.get("linestyle", "none")
        fig = viz.probplot(
            self.data,
            ax=ax,
            axtype=axtype,
            yscale=yscale,
            bestfit=bestfit,
            scatter_kws=scatter_kws,
        )

        if yscale == "log":
            pass

        if clearYLabels:
            ax.set_yticklabels([])

        if rotateticklabels:
            viz.rotate_tick_labels(ax, 45, "x", ha="right")

        if bestfit:
            utils.fit_line()

        return fig

    def statplot(
        self,
        pos=1,
        yscale="log",
        shownotches=True,
        showmean=True,
        width=0.8,
        bacteria=False,
        ylabel=None,
        xlabel=None,
        axtype="prob",
        patch_artist=False,
        **plotopts,
    ):
        """Creates a two-axis figure with a boxplot & probability plot.

        Parameters
        ----------
        pos : int, optional (default=1)
            Location along x-axis where boxplot will be placed.
        yscale : string, optional ['linear' or 'log' (default)]
            Scale formatting of the y-axis
        shownotches : bool, optional (default=True)
            Toggles drawing of bootstrapped confidence interval around
            the median.
        showmean : bool, optional (default=True)
            Toggles plotting the mean value on the boxplot as a point.
            See also the `bacteria` kwarg
        width : float, optional (default=0.8)
            Width of boxplot on the axes (data units)
        bacteria : bool, optional (default False)
            If True, uses the geometric mean when `showmean` is True.
            Otherwise, the arithmetic mean is used.
        ylabel : string, optional or None (default):
            Label for y-axis
        xlabel : string or None (default):
            Label for x-axis of boxplot. If None, uses self.name
        axtype : string, optional (default = 'pp')
            Type of probability plot to be created. Options are:
                - 'prob': probabilty plot
                - 'pp': percentile plot
                - 'qq': quantile plot
        patch_artist : bool, optional (default = False)
            Toggles the use of patch artist instead of a line artists
            for the boxes
        plotopts : keyword arguments
            Additional options passed directly to `pyplot.plot` when
            drawing the data points on the probability plots.

        Returns
        -------
        fig : matplotlib Figure

        """

        # setup the figure and axes
        fig = pyplot.figure(figsize=(6.40, 3.00), facecolor="none", edgecolor="none")
        ax1 = pyplot.subplot2grid((1, 4), (0, 0))
        ax2 = pyplot.subplot2grid((1, 4), (0, 1), colspan=3)

        self.boxplot(
            ax=ax1,
            pos=pos,
            yscale=yscale,
            shownotches=shownotches,
            showmean=showmean,
            width=width,
            bacteria=bacteria,
            ylabel=ylabel,
            xlabel=xlabel,
            patch_artist=patch_artist,
            xlims={"left": pos - (0.6 * width), "right": pos + (0.6 * width)},
        )

        self.probplot(
            ax=ax2, yscale=yscale, axtype=axtype, ylabel=None, clearYLabels=True, **plotopts
        )

        ax1.yaxis.tick_left()
        ax2.yaxis.tick_right()
        fig.subplots_adjust(wspace=0.05)
        return fig

    def verticalScatter(
        self, ax=None, pos=1, ylabel=None, yscale="log", ignoreROS=True, markersize=6
    ):
        """Draws a clustered & jittered scatter plot of the data

        Parameters
        ----------
        ax : matplotlib axes object, optional or None (default)
            Axes on which the points with be drawn. If None, one will
            be created.
        pos : int, optional (default=1)
            Location along x-axis where data will be centered.
        jitter : float, optional (default=0.80)
            Width of the random x-values uniform distributed around
            `pos`
        alpha : float, optional (default=0.75)
            Opacity of the marker (1.0 -> opaque; 0.0 -> transparent)
        yscale : string, optional ['linear' or 'log' (default)]
            Scale formatting of the y-axis
        ylabel : string, optional or None (default):
            Label for y-axis
        ignoreROS : bool, optional (default = True)
            By default, this function will plot the original, non-ROS'd
            data with a different symbol for non-detects. If `True`, the
            and `self.useros` is `True`, this function will plot the
            ROS'd data with a single marker.
        markersize : int, optional (default = 6)
            Size of data markers on the figure in points.

        Returns
        -------
        fig : matplotlib Figure

        """

        fig, ax = validate.axes(ax)

        if not ignoreROS and self.useros:
            rescol = "final"
            hue_column = None
            hue_order = None
            df = self.dataframe.copy()
        else:
            rescol = self.rescol
            hue_column = "Censored"
            hue_order = [True, False]
            df = self.raw_data.copy()

        ax = (
            df.assign(pos=pos)
            .assign(Result=lambda df: df[rescol])
            .assign(Censored=lambda df: df[self.cencol])
            .pipe(
                (seaborn.swarmplot, "data"),
                y="Result",
                x="pos",
                hue=hue_column,
                hue_order=hue_order,
                size=markersize,
                marker=self.plot_marker,
                color=self.color,
                edgecolor="k",
                linewidth=1.25,
                ax=ax,
            )
        )

        ax.set_yscale(yscale)
        if ylabel is not None:
            ax.set_ylabel(ylabel)

        return fig


class Dataset:
    """Dataset: object for comparings two Location objects

    Parameters
    ----------
    influent, effluent : wqio.Location
        Location objects that will be compared. Data from each Location
        should be joinable on the dataframe's index.
    useros : bool (default = True)
        Toggles the use of Regression On Order Statistics to
        estimate non-detect values when computing statistics.
    name : string optional
        Name for the dataset.

    Notes
    -----
    Currently moving away from this in favor of DataCollections.

    """

    # TODO: constructor should take dataframe, and build Location object,
    # not the other way around. This will allow Dataset.influent = None
    # by passing in a dataframe where df.shape[0] == 0
    def __init__(self, influent, effluent, useros=True, name=None):
        # basic attributes
        self.influent = influent
        self.effluent = effluent
        self._name = name
        self._include = None
        self.useros = useros
        self._definition = {}
        self._cache = {}

    @cache_readonly
    def data(self):
        if self.effluent.hasData:
            effl = self.effluent.raw_data.copy()
        else:
            raise ValueError("effluent must have data")

        if self.influent.hasData:
            infl = self.influent.raw_data.copy()
        else:
            infl = pandas.DataFrame(
                index=self.effluent.raw_data.index,
                columns=self.effluent.raw_data.columns,
            )

        infl = utils.add_column_level(infl, "inflow", "station")
        effl = utils.add_column_level(effl, "outflow", "station")
        return infl.join(effl, how="outer")

    @cache_readonly
    def paired_data(self):
        if self.data is not None:
            return self.data.dropna()

    @cache_readonly
    def n_pairs(self):
        if self.paired_data is not None:
            return self.paired_data.shape[0]
        else:
            return 0

    @cache_readonly
    def _non_paired_stats(self):
        return self.influent.data is not None and self.effluent.data is not None

    @cache_readonly
    def _paired_stats(self):
        return self._non_paired_stats and self.paired_data.shape[0] > 20

    def __repr__(self):
        x = f"<wqio.Dataset>\n  N influent  {self.influent.N}\n  N effluent = {self.effluent.N}"
        if self.definition is not None:
            for k, v in self.definition.items():
                x = f"{x}\n  {k.title()} = {v}"
        return x

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, value):
        self._definition = value

    @property
    def include(self):
        if self._include is None:
            self._include = self.influent.include and self.effluent.include
        return self._include

    @include.setter
    def include(self, value):
        self._include = value

    @property
    def exclude(self):
        return not self.include

    # stats describing the dataset
    @cache_readonly
    def medianCIsOverlap(self):
        overlap = True
        if self.influent.hasData and self.effluent.hasData:
            overlap = utils.check_interval_overlap(
                self.influent.median_conf_interval,
                self.effluent.median_conf_interval,
                oneway=False,
            )
        return overlap

    @cache_readonly
    def wilcoxon_z(self):
        """The Wilcoxon Z-statistic.

        Tests the null hypothesis that the influent and effluent data are
        sampled from the same statistical distribution.

        Notes
        -----
        Operates on natural log-transormed paired data only.

        See also
        --------
        scipy.stats.wilcoxon

        """
        if self._wilcoxon_stats is not None:
            return self._wilcoxon_stats[0]

    @cache_readonly
    def wilcoxon_p(self):
        """Two-sided p-value of the Wilcoxon test

        See also
        --------
        scipy.stats.wilcoxon

        """
        if self._wilcoxon_stats is not None:
            return self._wilcoxon_stats[1]

    @cache_readonly
    def mannwhitney_u(self):
        """Mann-Whitney U-statistic.

        Performs a basic rank-sum test.

        Notes
        -----
        Operates on untransformed, non-paired data.

        See also
        --------
        scipy.stats.mannwhitneyu

        """
        if self._mannwhitney_stats is not None:
            return self._mannwhitney_stats[0]

    @cache_readonly
    def mannwhitney_p(self):
        """Two-sided p-value of the Mann-Whitney test

        Notes
        -----
        Scipy functions returns the 1-sided p-values. This corrects for that.

        See also
        --------
        scipy.stats.mannwhitneyu

        """
        if self._mannwhitney_stats is not None:
            return self._mannwhitney_stats[1]

    @cache_readonly
    def kendall_tau(self):
        """The Kendall-Tau statistic.

        Measure of the correspondence between two the rankings of influent
        and effluent data.

        Notes
        -----
        Operates on paired data only.

        See also
        --------
        scipy.stats.kendalltau

        """
        if self._kendall_stats is not None:
            return self._kendall_stats[0]

    @cache_readonly
    def kendall_p(self):
        """Two-sided p-value of the Kendall test

        See also
        --------
        scipy.stats.kendalltau

        """
        if self._kendall_stats is not None:
            return self._kendall_stats[1]

    @cache_readonly
    def spearman_rho(self):
        """The Spearman's rho statistic.

        Tests for monotonicity of the relationship between influent and.
        effluent data.

        Notes
        -----
        Operates on paired data only.

        See also
        --------
        scipy.stats.spearmanr

        """
        if self._spearman_stats is not None:
            return self._spearman_stats[0]

    @cache_readonly
    def spearman_p(self):
        """Two-sided p-value of the Spearman test

        See also
        --------
        scipy.stats.spearmanr

        """
        if self._spearman_stats is not None:
            return self._spearman_stats[1]

    @cache_readonly
    def ttest_t(self):
        return self._ttest_stats[0]

    @cache_readonly
    def ttest_p(self):
        return self._ttest_stats[1]

    @cache_readonly
    def levene_ks(self):
        return self._levene_stats[0]

    @cache_readonly
    def levene_p(self):
        return self._levene_stats[1]

    @cache_readonly
    def theil_medslope(self):
        return self._theil_stats["medslope"]

    @cache_readonly
    def theil_intercept(self):
        return self._theil_stats["intercept"]

    @cache_readonly
    def theil_loslope(self):
        return self._theil_stats["loslope"]

    @cache_readonly
    def theil_hislope(self):
        return self._theil_stats["hislope"]

    # helper objects for the stats
    @cache_readonly
    def _wilcoxon_stats(self):
        if self._paired_stats:
            return stats.wilcoxon(
                numpy.log(self.paired_data.inflow.res),
                numpy.log(self.paired_data.outflow.res),
            )

    @cache_readonly
    def _mannwhitney_stats(self):
        if self._non_paired_stats:
            return stats.mannwhitneyu(
                self.influent.data, self.effluent.data, alternative="two-sided"
            )

    @cache_readonly
    def _kendall_stats(self):
        if self._paired_stats:
            return stats.kendalltau(self.paired_data.inflow.res, self.paired_data.outflow.res)

    @cache_readonly
    def _spearman_stats(self):
        if self._paired_stats:
            return stats.spearmanr(
                self.paired_data.inflow.res.values, self.paired_data.outflow.res.values
            )

    @cache_readonly
    def _ttest_stats(self):
        if self._non_paired_stats:
            return stats.ttest_ind(self.influent.data, self.effluent.data, False)

    @cache_readonly
    def _levene_stats(self):
        if self._non_paired_stats:
            return stats.levene(self.influent.data, self.effluent.data, center="median")

    @cache_readonly
    def _theil_stats(self):
        return self.theilSlopes()

    def theilSlopes(self, log_infl=False, log_effl=False):
        output = None
        # influent data
        infl = self.paired_data.inflow.res.values
        if log_infl:
            infl = numpy.log(infl)

        # effluent data
        effl = self.paired_data.outflow.res.values
        if log_effl:
            effl = numpy.log(effl)

        if self.influent.fractionND <= 0.5 and self.effluent.fractionND <= 0.5:
            # we need to make sure that the "y" values are the
            # Location with the greatest NUnique to avoid a
            # slope of zero if possible
            if self.influent.NUnique <= self.effluent.NUnique:
                inverted = False
                theilstats = stats.mstats.theilslopes(effl, x=infl)
            else:
                inverted = True
                theilstats = stats.mstats.theilslopes(infl, x=effl)

            # stuff things into a dictionary
            if not inverted:
                output = {
                    "medslope": theilstats[0],
                    "intercept": theilstats[1],
                    "loslope": theilstats[2],
                    "hislope": theilstats[3],
                    "is_inverted": inverted,
                }
            else:
                output = {
                    "medslope": 1 / theilstats[0],
                    "intercept": -1 * theilstats[1] / theilstats[0],
                    "loslope": 1 / theilstats[2],
                    "hislope": 1 / theilstats[3],
                    "is_inverted": inverted,
                }

            output["estimated_effluent"] = _estimate_from_fit(
                infl,
                output["medslope"],
                output["intercept"],
                xlog=log_infl,
                ylog=log_effl,
            )

            output["estimate_error"] = (
                self.paired_data.outflow.res.values - output["estimated_effluent"]
            )

        return output

    # plotting methods
    def boxplot(
        self,
        ax=None,
        pos=1,
        yscale="log",
        shownotches=True,
        showmean=True,
        width=0.8,
        bacteria=False,
        ylabel=None,
        xlims=None,
        bothTicks=True,
        offset=0.5,
        patch_artist=False,
    ):
        """Adds a boxplot to a matplotlib figure

        Parameters
        ----------
        ax : matplotlib axes object or None (default), optional
            Axes on which the boxplot with be drawn. If None, one will
            be created.
        pos : int, optional (default=1)
            Location along x-axis where boxplot will be placed.
        yscale : optional string ['linear' or 'log' (default)]
            Scale formatting of the y-axis
        shownotches : bool, optional (default=True)
            Toggles drawing of bootstrapped confidence interval around
            the median.
        showmean : bool, optional (default is True)
            Toggles plotting the mean value on the boxplot as a point.
            See also the `bacteria` kwarg
        width : float, optional (default=0.8)
            Width of boxplot on the axes (data units)
        bacteria : bool, optional (default is False)
            If True, uses the geometric mean when `showmean` is True.
            Otherwise, the arithmetic mean is used.
        ylabel : string or None (default):
            Label for y-axis
        xlims : sequence (length=2), dict, or None (default), optional
            Custom limits of the x-axis. If None, defaults to
            [pos-1, pos+1].
        bothTicks : bool, optional (default is True)
            If True, each box gets a tick label. Otherwise, both get a
            single tick.
        offset : float, optional (default = 0.5)
            Spacing, in x-axis data coordinates, of the boxplots.
        patch_artist : bool, optional (default = False)
            Toggles the use of patch artist instead of a line artists
            for the boxes

        Returns
        -------
        fig : matplotlib Figure

        """

        fig, ax = validate.axes(ax)

        for loc, offset in zip([self.influent, self.effluent], [-1 * offset, offset]):
            if loc is not None:
                y_log = yscale == "log"
                bxpstats = loc.boxplot_stats(log=y_log, bacteria=bacteria)
                viz.boxplot(
                    bxpstats,
                    ax=ax,
                    position=pos + offset,
                    width=width,
                    color=loc.color,
                    marker=loc.plot_marker,
                    patch_artist=patch_artist,
                    showmean=showmean,
                    shownotches=shownotches,
                )

        ax.set_yscale(yscale)
        if y_log:
            ax.yaxis.set_major_formatter(viz.log_formatter(use_1x=False))

        if ylabel:
            ax.set_ylabel(ylabel)

        if xlims is None:
            ax.set_xlim([pos - 1, pos + 1])
        else:
            if isinstance(xlims, dict):
                ax.set_xlim(**xlims)
            else:
                ax.set_xlim(xlims)

        if bothTicks:
            ax.set_xticks([pos - offset, pos + offset])
            ax.set_xticklabels([self.influent.name, self.effluent.name])
        else:
            ax.set_xticks([pos])
            if self.name is not None:
                ax.set_xticklabels([self.name])
            else:
                ax.set_xticklabels([""])

        return fig

    def probplot(
        self,
        ax=None,
        yscale="log",
        axtype="prob",
        ylabel=None,
        clearYLabels=False,
        rotateticklabels=True,
        bestfit=False,
    ):
        """Adds probability plots to a matplotlib figure

        Parameters
        ----------
        ax : matplotlib axes object, optional or None (default).
            The Axes on which to plot. If None is provided, one will be
            created.
        yscale : string , optional(default = 'log')
            Scale for the y-axis. Use 'log' for logarithmic (default) or
            'linear'.
        axtype : string, optional (default = 'pp')
            Type of plot to be created. Options are:
                - 'prob': probabilty plot
                - 'pp': percentile plot
                - 'qq': quantile plot
        ylabel : string, optional or None (default)
            Axis label for the plot.
        clearYLabels : bool, optional (default is False)
            If True, removed y-*tick* labels from teh Axes.
        rotateticklabels : bool, optional (default is True)
            If True, the tick labels of the probability axes will be
            rotated.
        bestfit : bool, optional (default is False)
            Specifies whether a best-fit line should be added to the
            plot.

        Returns
        -------
        fig : matplotlib Figure

        """

        fig, ax = validate.axes(ax)

        for loc in [self.influent, self.effluent]:
            if loc.include:
                loc.probplot(
                    ax=ax,
                    clearYLabels=clearYLabels,
                    axtype=axtype,
                    yscale=yscale,
                    bestfit=bestfit,
                    rotateticklabels=rotateticklabels,
                )

        xlabels = {
            "pp": "Theoretical percentiles",
            "qq": "Theoretical quantiles",
            "prob": r"Non-exceedance probability (%)",
        }

        ax.set_xlabel(xlabels[axtype])
        ax.legend(loc="lower right", frameon=True)

        if ylabel is not None:
            ax.set_ylabel(ylabel)

        if rotateticklabels:
            viz.rotate_tick_labels(ax, 45, "x")

        return fig

    def statplot(
        self,
        pos=1,
        yscale="log",
        shownotches=True,
        showmean=True,
        width=0.8,
        bacteria=False,
        ylabel=None,
        axtype="qq",
        patch_artist=False,
    ):
        """Creates a two-axis figure with a boxplot & probability plot.

        Parameters
        ----------
        pos : int, optional (default=1)
            Location along x-axis where boxplot will be placed.
        yscale : string, optional ['linear' or 'log' (default)]
            Scale formatting of the y-axis
        shownotches : bool, optional (default=True)
            Toggles drawing of bootstrapped confidence interval around
            the median.
        showmean : bool, optional (default=True)
            Toggles plotting the mean value on the boxplot as a point.
            See also the `bacteria` kwarg
        width : float, optional (default=0.8)
            Width of boxplot on the axes (data units)
        bacteria : bool, optional (default False)
            If True, uses the geometric mean when `showmean` is True.
            Otherwise, the arithmetic mean is used.
        ylabel : string, optional or None (default):
            Label for y-axis
        probAxis : bool, optional (default = True)
            Toggles the display of probabilities (True) or Z-scores
            (i.e., theoretical quantiles) on the x-axis
        patch_artist : bool, optional (default = False)
            Toggles the use of patch artist instead of a line artists
            for the boxes

        Returns
        -------
        fig : matplotlib Figure

        """

        # setup the figure and axes
        fig = pyplot.figure(figsize=(6.40, 3.00), facecolor="none", edgecolor="none")
        ax1 = pyplot.subplot2grid((1, 4), (0, 0))
        ax2 = pyplot.subplot2grid((1, 4), (0, 1), colspan=3)

        self.boxplot(
            ax=ax1,
            pos=pos,
            yscale=yscale,
            shownotches=shownotches,
            showmean=showmean,
            width=width,
            bacteria=bacteria,
            ylabel=ylabel,
            patch_artist=patch_artist,
        )

        self.probplot(
            ax=ax2,
            yscale=yscale,
            axtype=axtype,
            ylabel=None,
            clearYLabels=True,
            rotateticklabels=True,
        )

        ax1.yaxis.tick_left()
        ax2.yaxis.tick_right()
        fig.subplots_adjust(wspace=0.05)
        return fig

    def jointplot(self, hist=False, kde=True, rug=True, **scatter_kws):
        """Create a joint distribution plot for the dataset

        Parameters
        ----------
        hist : bool, optional (default is False)
            Toggles showing histograms on the distribution plots
        kde : bool, optional (default is True)
            Toggles showing KDE plots on the distribution plots
        rug : bool, optional (default is True)
            Toggles showing rug plots on the distribution plots
        **scatter_kws : keyword arguments
            Optionals passed directly to Dataset.scatterplot

        Returns
        -------
        jg : seaborn.JointGrid

        See also
        --------
        seaborn.JointGrid
        seaborn.jointplot
        seaborn.displot
        Dataset.scatterplot

        """

        showlegend = scatter_kws.pop("showlegend", True)
        _ = scatter_kws.pop("ax", None)

        if self.paired_data is not None:
            data = self.paired_data.xs("res", level="quantity", axis=1)
            jg = seaborn.JointGrid(x="inflow", y="outflow", data=data)
            self.scatterplot(ax=jg.ax_joint, showlegend=False, **scatter_kws)
            jg.plot_marginals(seaborn.displot, rug=rug, kde=kde)

            jg.ax_marg_x.set_xscale(scatter_kws.pop("xscale", "log"))
            jg.ax_marg_y.set_yscale(scatter_kws.pop("yscale", "log"))

            if showlegend:
                jg.ax_joint.legend(loc="upper left")

        return jg

    def scatterplot(
        self,
        ax=None,
        xscale="log",
        yscale="log",
        showlegend=True,
        xlabel=None,
        ylabel=None,
        one2one=False,
        useros=False,
        bestfit=False,
        minpoints=3,
        eqn_pos="lower right",
        equal_scales=True,
        fitopts=None,
        **markeropts,
    ):
        """Creates an influent/effluent scatter plot

        Parameters
        ----------
        ax : matplotlib axes object or None (default), optional
            Axes on which the scatterplot with be drawn. If None, one
            will be created.
        xscale, yscale : string ['linear' or 'log' (default)], optional
            Scale formatting of the [x|y]-axis
        xlabel, ylabel : string or None (default), optional:
            Label for [x|y]-axis. If None, will be 'Influent'. If the
            dataset definition is available and incldues a Parameter,
            that will be included as will. For no label, use
            e.g., `xlabel = ""`.
        showlegend : bool, optional (default is True)
            Toggles including a legend on the plot.
        one2one : bool, optional (default is False), optional
            Toggles the inclusion of the 1:1 line (i.e. line of
            equality).
        useros : bool, optional (default is False)
            Toggles the use of the ROS'd results. If False, raw results
            (i.e., detection limit for NDs) are used with varying
            symbology.

        Returns
        ------
        fig : matplotlib Figure

        """

        # set up the figure/axes
        fig, ax = validate.axes(ax)

        # set the scales
        ax.set_xscale(xscale)
        ax.set_yscale(yscale)

        # common symbology
        commonopts = dict(linestyle="none", markeredgewidth=0.5, markersize=6, zorder=10)

        # plot the ROSd'd result, if requested
        if useros:
            raise ValueError

        # plot the raw results, if requested
        else:
            plot_params = [
                dict(
                    label="Detected data pairs",
                    which="neither",
                    marker="o",
                    alpha=0.8,
                    markerfacecolor="black",
                    markeredgecolor="white",
                ),
                dict(
                    label="Influent not detected",
                    which="influent",
                    marker="v",
                    alpha=0.45,
                    markerfacecolor="none",
                    markeredgecolor="black",
                ),
                dict(
                    label="Effluent not detected",
                    which="effluent",
                    marker="<",
                    alpha=0.45,
                    markerfacecolor="none",
                    markeredgecolor="black",
                ),
                dict(
                    label="Both not detected",
                    which="both",
                    marker="d",
                    alpha=0.45,
                    markerfacecolor="none",
                    markeredgecolor="black",
                ),
            ]
            for pp in plot_params:
                kwargs = {**commonopts, **pp, **markeropts}
                self._plot_nds(ax, **kwargs)

        if xscale == "linear":
            ax.set_xlim(left=0)

        if yscale == "linear":
            ax.set_ylim(bottom=0)

        # unify the axes limits
        if xscale == yscale and equal_scales:
            ax.set_aspect("equal")
            axis_limits = [
                numpy.min([ax.get_xlim(), ax.get_ylim()]),
                numpy.max([ax.get_xlim(), ax.get_ylim()]),
            ]
            ax.set_ylim(axis_limits)
            ax.set_xlim(axis_limits)

        elif yscale == "linear" or xscale == "linear":
            axis_limits = [
                numpy.min([numpy.min(ax.get_xlim()), numpy.min(ax.get_ylim())]),
                numpy.max([ax.get_xlim(), ax.get_ylim()]),
            ]

        # include the line of equality, if requested
        if one2one:
            viz.one2one_line(
                ax,
                linestyle="-",
                linewidth=1.25,
                alpha=0.50,
                color="black",
                zorder=5,
                label="1:1 line",
            )

        detects = self.paired_data.loc[
            (~self.paired_data[("inflow", "qual")].isin(self.influent.ndvals))
            & (~self.paired_data[("outflow", "qual")].isin(self.effluent.ndvals))
        ].xs("res", level=1, axis=1)

        if bestfit and detects.shape[0] >= minpoints:
            if xscale == "log" and yscale == "log":
                fitlogs = "both"
            elif xscale == "log":
                fitlogs = "x"
            elif yscale == "log":
                fitlogs = "y"
            else:
                fitlogs = None

            x = detects["inflow"]
            y = detects["outflow"]

            fitopts = validate.at_least_empty_dict(fitopts, fitlogs=fitlogs)
            xhat, yhat, modelres = utils.fit_line(x, y, **fitopts)

            ax.plot(xhat, yhat, "k--", alpha=0.75, label="Best-fit")

            if eqn_pos is not None:
                positions = {
                    "lower left": (0.05, 0.15),
                    "lower right": (0.59, 0.15),
                    "upper left": (0.05, 0.95),
                    "upper right": (0.59, 0.95),
                }
                vert_offset = 0.1
                try:
                    txt_x, txt_y = positions.get(eqn_pos.lower())
                except KeyError:
                    raise ValueError(f"`eqn_pos` must be on of {list.positions.keys()}")
                # annotate axes with stats

                ax.annotate(
                    r"$\log(y) = {} \, \log(x) + {}$".format(
                        utils.sig_figs(modelres.params[1], n=3),
                        utils.sig_figs(modelres.params[0], n=3),
                    ),
                    (txt_x, txt_y),
                    xycoords="axes fraction",
                )

                ax.annotate(
                    "Slope p-value: {}\nIntercept p-value: {}".format(
                        utils.process_p_vals(modelres.pvalues[1]),
                        utils.process_p_vals(modelres.pvalues[0]),
                    ),
                    (txt_x, txt_y - vert_offset),
                    xycoords="axes fraction",
                )

        # setup the axes labels
        if xlabel is None:
            xlabel = "Influent"

        if ylabel is None:
            ylabel = "Effluent"

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        # show legend, if requested
        if showlegend:
            leg = ax.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.00)
            leg.get_frame().set_alpha(0.00)
            leg.get_frame().set_edgecolor("none")

        return fig

    def _plot_nds(self, ax, which="both", label="_no_legend", **markerkwargs):
        """
        Helper function for scatter plots -- plots various combinations
        of non-detect paired data
        """
        i_nondetect = self.paired_data[("inflow", "cen")]
        o_nondetect = self.paired_data[("outflow", "cen")]

        i_detect = ~self.paired_data[("inflow", "cen")]
        o_detect = ~self.paired_data[("outflow", "cen")]

        index_combos = {
            "both": i_nondetect & o_nondetect,
            "influent": i_nondetect & o_detect,
            "effluent": i_detect & o_nondetect,
            "neither": i_detect & o_detect,
        }

        try:
            index = index_combos[which]
        except KeyError:
            msg = '`which` must be "both", "influent", ' '"effluent", or "neighter"'
            raise ValueError(msg)

        x = self.paired_data.loc[index][("inflow", "res")]
        y = self.paired_data.loc[index][("outflow", "res")]
        return ax.plot(x, y, label=label, **markerkwargs)



import tempfile

import pytest

from wqio import datasets


@pytest.mark.parametrize("fname", ["bmpdata", "nsqd"])
def test_download(fname):
    with tempfile.TemporaryDirectory() as tmpdir:
        dstpath = datasets.download(fname, data_dir=tmpdir, year=2016)
        assert dstpath.exists()
        assert dstpath.name == fname + ".csv"



import logging

import numpy
import pandas

from wqio import validate
from wqio.utils import misc

_logger = logging.getLogger(__name__)


def get_season(date):
    """Defines the season from a given date.

    Parameters
    ----------
    date : datetime.datetime object or similar
        Any object that represents a date and has `.month` and `.day`
        attributes

    Returns
    -------
    season : str

    Notes
    -----
    Assumes that all seasons changed on the 22nd (e.g., all winters
    start on Decemeber 22). This isn't strictly true, but it's good
    enough for now.

    """

    date = validate.timestamp(date)
    day = date.dayofyear
    leap_year = int(date.is_leap_year)

    spring = numpy.arange(80, 172) + leap_year
    summer = numpy.arange(172, 264) + leap_year
    autumn = numpy.arange(264, 355) + leap_year

    if day in spring:
        season = "spring"
    elif day in summer:
        season = "summer"
    elif day in autumn:
        season = "autumn"
    else:
        season = "winter"

    return season


def make_timestamp(row, datecol="sampledate", timecol="sampletime", issuewarnings=False):
    """Makes a pandas.Timestamp from separate date/time columns

    Parameters
    ----------
    row : dict-like (ideallty a row in a dataframe)
    datecol : string (default = 'sampledate')
        Name of the column containing the dates
    timecol : string (default = 'sampletime')
        Name of the column containing the times
    issuewarnings : bool (default = False)
        Toggles UserWarnings when the fallback date (1901-01-01) or
        fallback time (00:00) must be used.

    Returns
    -------
    tstamp : pandas.Timestamp

    """

    fallback_datetime = pandas.Timestamp("1901-01-01 00:00")

    if row[datecol] is None or pandas.isnull(row[datecol]):
        fb_date = True
    else:
        fb_date = False
        try:
            date = pandas.Timestamp(row[datecol]).date()
        except ValueError:
            fb_date = True

    if fb_date:
        date = fallback_datetime.date()
        if issuewarnings:  # pragma: no cover
            misc.log_or_warn(
                f"Using fallback date from {row[datecol]}",
                UserWarning if issuewarnings else None,
                logger=_logger,
            )

    if row[timecol] is None or pandas.isnull(row[timecol]):
        fb_time = True
    else:
        fb_time = False
        try:
            time = pandas.Timestamp(row[timecol]).time()
        except ValueError:
            fb_time = True

    if fb_time:
        time = fallback_datetime.time()
        if issuewarnings:  # pragma: no cover
            misc.log_or_warn(
                f"Using fallback time from {row[timecol]}",
                UserWarning if issuewarnings else None,
                logger=_logger,
            )

    dtstring = f"{date} {time}"
    tstamp = pandas.Timestamp(dtstring)

    return tstamp


def get_wateryear(date):
    """Returns the water year of a given date

    Parameters
    ----------
    date : datetime-like
        A datetime or Timestamp object

    Returns
    -------
    wateryear : string
        The water year of `date`

    Example
    -------
    >>> import datetime
    >>> import wqio
    >>> x = datetime.datetime(2005, 11, 2)
    >>> print(wqio.utils.get_wateryear(x))
    2005/2006

    """

    year = date.year
    yearstring = "{}/{}"
    if date.month >= 10:
        return yearstring.format(year, year + 1)
    else:
        return yearstring.format(year - 1, year)
